"""
spymicmac.micmac is a collection of tools for interfacing with MicMac
"""
import os
import subprocess
import shutil
import numpy as np
from osgeo import gdal
import pandas as pd
import lxml.etree as etree
import lxml.builder as builder
import difflib
import xml.etree.ElementTree as ET
from glob import glob
from scipy.interpolate import LinearNDInterpolator
from shapely.geometry.point import Point
from shapely.geometry import LineString
from shapely.strtree import STRtree
from skimage.io import imread
from skimage.measure import ransac
from skimage.transform import AffineTransform
from pybob.bob_tools import mkdir_p
from pybob.GeoImg import GeoImg
from pybob.ddem_tools import nmad
from spymicmac.usgs import get_usgs_footprints


######################################################################################################################
# MicMac interfaces - write xml files for MicMac to read
######################################################################################################################
def write_neighbour_images(imlist, fprints=None, nameField='ID', prefix='OIS-Reech_', fileExt='.tif',
                           dataset='AERIAL_COMBIN'):
    """
    Using a list of images and a collection of image footprints, return a list of potential image pairs for processing
    with Tapioca.

    :param list imlist: a list of (original) image names to use (e.g., without 'OIS-Reech\_')
    :param GeoDataFrame fprints: a vector dataset of footprint polygons. If not provided, will attempt to download
        metadata from USGS for the images.
    :param str nameField: the field in fprints table that contains the image name
    :param str prefix: the prefix attached to the image name read by Tapioca (default: 'OIS-Reech\_')
    :param str fileExt: the file extension for the images read by Tapioca (default: .tif)
    :param dataset: the USGS dataset name to search if no footprints are provided (default: AERIAL_COMBIN)
    """
    E = builder.ElementMaker()
    NamedRel = E.SauvegardeNamedRel()

    if fprints is None:
        fprints = get_usgs_footprints(imlist, dataset=dataset)
    else:
        fprints = fprints[fprints[nameField].isin(imlist)]

    s = STRtree([f for f in fprints['geometry'].values])

    for i, row in fprints.iterrows():
        fn = row[nameField]
        fp = row['geometry']

        print(fn)

        res = s.query(fp)
        intersects = [c for c in res if fp.intersection(c).area > 0]
        fnames = [fprints[nameField][fprints['geometry'] == c].values[0] for c in intersects]
        try:
            fnames.remove(fn)
        except ValueError:
            pass

        for f in fnames:
            this_pair = E.Cple(' '.join([prefix + fn + fileExt,
                                         prefix + f + fileExt]))
            NamedRel.append(this_pair)

    tree = etree.ElementTree(NamedRel)
    tree.write('FileImagesNeighbour.xml', pretty_print=True, xml_declaration=True, encoding="utf-8")


def get_gcp_meas(im_name, meas_name, in_dir, E, nodist=None, gcp_name='GCP'):
    """
    Create an lxml.builder.ElementMaker object with a GCP name and the image (row, pixel) location.

    :param str im_name: the image name to write the GCP location for.
    :param str meas_name: the name of the file to read the point locations from.
    :param str in_dir: the name of the directory where the images and measures files are located.
    :param lxml.builder.ElementMaker E: an ElementMaker object for writing to the xml file.
    :param str nodist: the name of the directory
    :param str gcp_name: the prefix (e.g., GCP0, GCP1, etc.) for the GCP name (default: GCP).
    :return:
        - **this_im_meas** (*lxml.builder.ElementMaker*) -- an ElementMaker object with the GCP location in the image.
    """
    im = gdal.Open(os.path.sep.join([in_dir, im_name]))
    maxj = im.RasterXSize
    maxi = im.RasterYSize

    impts = pd.read_csv(os.path.join(in_dir, meas_name), sep=' ', names=['j', 'i'])
    if nodist is not None:
        impts_nodist = pd.read_csv(os.path.join(in_dir, nodist), sep=' ', names=['j', 'i'])

    this_im_mes = E.MesureAppuiFlottant1Im(E.NameIm(im_name))
    for ind, row in impts.iterrows():
        in_im = 0 < row.j < maxj and 0 < row.i < maxi
        if nodist is not None:
            in_nd = -200 < impts_nodist.j[ind]+200 < maxj and -200 < impts_nodist.i[ind] < maxi+200
            in_im = in_im and in_nd
        if in_im:
            this_mes = E.OneMesureAF1I(
                                E.NamePt('{}{}'.format(gcp_name, ind+1)),
                                E.PtIm('{} {}'.format(row['j'], row['i']))
                            )
            this_im_mes.append(this_mes)
    return this_im_mes


def get_im_meas(gcps, E):
    """
    Populate an lxml.builder.ElementMaker object with GCP image locations, for writing to xml files.

    :param pandas.DataFrame gcps: a DataFrame with the GCPs to find image locations for.
    :param lxml.builder.ElementMaker E: an ElementMaker object for writing to the xml file.
    :return:
        - **pt_els** (*list*) -- a list of ElementMaker objects corresponding to each GCP image location.
    """
    pt_els = []
    for ind, row in gcps.iterrows():
        this_mes = E.OneMesureAF1I(
                        E.NamePt(row['gcp']),
                        E.PtIm('{} {}'.format(row['im_col'], row['im_row']))
                        )
        pt_els.append(this_mes)
    return pt_els


def parse_im_meas(fn_meas):
    """
    Read an xml file with GCP image locations into a pandas DataFrame.

    :param fn_meas: the name of the measures file to read.
    :return:
        - **gcp_df** (*pandas.DataFrame*) -- a DataFrame with gcp names and image locations.
    """
    gcp_df = pd.DataFrame()
    root = ET.parse(fn_meas).getroot()
    measures = root.findall('MesureAppuiFlottant1Im')[0]
    for i, mes in enumerate(measures.findall('OneMesureAF1I')):
        gcp_df.loc[i, 'name'] = mes.find('NamePt').text
        pt = mes.find('PtIm').text.split()
        gcp_df.loc[i, 'i'] = float(pt[1])
        gcp_df.loc[i, 'j'] = float(pt[0])
    return gcp_df


def generate_measures_files(joined=False):
    """
    Create id_fiducial.txt, MeasuresCamera.xml, and Tmp-SL-Glob.xml files for KH-9 Hexagon images.

    :param bool joined: generate files for joined scene (220x460 mm) instead of half (220x230mm)
    """
    i_list = np.arange(22, -1, -1)
    if not joined:
        j_list = np.arange(0, 24)
    else:
        j_list = np.arange(0, 47)

    J, I = np.meshgrid(np.arange(0, j_list.size), np.arange(0, i_list.size))
    gcp_names = list(zip(I[0, :], J[0, :]))
    for i in range(1, i_list.size):
        gcp_names.extend(list(zip(I[i, :], J[i, :])))

    JJ, II = np.meshgrid(np.round(j_list).astype(int), np.round(i_list).astype(int))
    ij = list(zip(II[0, :], JJ[0, :]))
    for i in np.arange(1, i_list.size):
        ij.extend(list(zip(II[i, :], JJ[i, :])))
    ij = 10 * np.array(ij)

    E = builder.ElementMaker()
    ImMes = E.MesureAppuiFlottant1Im(E.NameIm('Glob'))
    SpGlob = E.SetPointGlob()

    gcp_df = pd.DataFrame()
    with open('id_fiducial.txt', 'w') as f:
        for i, ind in enumerate(gcp_names):
            row, col = ind
            gcp_name = 'GCP_{}_{}'.format(row, col)
            gcp_df.loc[i, 'gcp'] = gcp_name
            gcp_df.loc[i, 'im_row'] = ij[i, 0]
            gcp_df.loc[i, 'im_col'] = ij[i, 1]

            pt_glob = E.PointGlob(E.Type('eNSM_Pts'),
                                  E.Name(gcp_name),
                                  E.LargeurFlou('0'),
                                  E.NumAuto('0'),
                                  E.SzRech('-1'))
            SpGlob.append(pt_glob)
            print(gcp_name, file=f)

    tree = etree.ElementTree(SpGlob)
    tree.write('Tmp-SL-Glob.xml', pretty_print=True, xml_declaration=True, encoding="utf-8")

    pt_els = get_im_meas(gcp_df, E)
    for p in pt_els:
        ImMes.append(p)

    outxml = E.SetOfMesureAppuisFlottants(ImMes)
    tree = etree.ElementTree(outxml)
    tree.write('MeasuresCamera.xml', pretty_print=True, xml_declaration=True, encoding="utf-8")

    with open('id_fiducial.txt', 'w') as f:
        for gcp in gcp_names:
            row, col = gcp
            print('GCP_{}_{}'.format(row, col), file=f)


def get_match_pattern(imlist):
    """
    Given a list of image names, return a match pattern that can be passed to MicMac command line functions.

    :param list imlist: a list of image names.
    :return:
        - **pattern** (*str*) -- a match pattern (e.g., "OIS.*tif") that can be passed to MicMac functions.
    """
    matches = []
    for i, this_im in enumerate(imlist[:-1]):
        for im in imlist[i + 1:]:
            matches.extend(list(difflib.SequenceMatcher(None, this_im, im).get_matching_blocks()))

    good_matches = set([m for m in matches if m.size > 0 and m.a == m.b])
    start_inds = set([m.a for m in good_matches])
    ind_lengths = [(ind, min([m.size for m in good_matches if m.a == ind])) for ind in start_inds]
    ind_lengths.sort()

    first, last = ind_lengths[0][1], ind_lengths[1][0]
    return imlist[0][:first] + '(' + '|'.join([im[first:last] for im in imlist]) + ')' + imlist[0][last:]


def write_auto_mesures(gcps, sub, outdir, outname='AutoMeasures'):
    """
    Write a file with GCP locations in relaive space (x, y, z) to use with get_autogcp_locations.sh

    :param pandas.DataFrame gcps: a DataFrame with the GCPs to save.
    :param str sub: the name of the block, if multiple blocks are being used (e.g., '_block1'). If not, use ''.
    :param str outdir: the output directory to save the files to.
    :param str outname: the base name of the file to create (default: AutoMeasures).
    """
    with open(os.path.join(outdir, '{}{}.txt'.format(outname, sub)), 'w') as f:
        for i, row in gcps.iterrows():
            print('{} {} {}'.format(row.rel_x, row.rel_y, row.el_rel), file=f)


def get_valid_image_points(shape, pts, pts_nodist):
    """
    Find which image points are located within an image based on the size of the image.

    :param shape: the shape of the image (rows, columns) to determine valid points for.
    :param pandas.DataFrame pts: a DataFrame containing point locations (i, j)
    :param pandas.DataFrame pts_nodist: a DataFrame containing point locations (i, j) calculated using no camera distortion.
    :return:
        - **valid_pts** (*array-like*) -- an array of the points that are located within the image shape.
    """
    maxi, maxj = shape

    in_im = np.logical_and.reduce((0 < pts.j, pts.j < maxj,
                                   0 < pts.i, pts.i < maxi))
    in_nd = np.logical_and.reduce((-200 < pts_nodist.j, pts_nodist.j < maxj + 200,
                                   -200 < pts_nodist.i, pts_nodist.i < maxi + 200))

    return np.logical_and(in_im, in_nd)


def write_image_mesures(imlist, gcps, outdir='.', sub='', ort_dir='Ortho-MEC-Relative'):
    """
    Create a Measures-S2D.xml file (row, pixel) for each GCP in each image from a list of image names.

    :param list imlist: a list of image names.
    :param pandas.DataFrame gcps: a DataFrame of GCPs.
    :param str outdir: the output directory to save the files to.
    :param str sub: the name of the block, if multiple blocks are being used (e.g., '_block1').
    :param str ort_dir: the Ortho-MEC directory where the images are located.
    """
    E = builder.ElementMaker()
    MesureSet = E.SetOfMesureAppuisFlottants()

    for im in imlist:
        print(im)
        img_geo = GeoImg(os.path.join(ort_dir, 'Ort_' + im))
        impts = pd.read_csv('Auto-{}.txt'.format(im), sep=' ', names=['j', 'i'])
        # impts_nodist = pd.read_csv('NoDist-{}.txt'.format(im), sep=' ', names=['j', 'i'])

        xmin, xmax, ymin, ymax = img_geo.find_valid_bbox()

        # valid_pts = get_valid_image_points(img.shape, impts, impts_nodist)
        valid_pts = np.logical_and.reduce([xmin <= gcps.rel_x, gcps.rel_x < xmax,
                                           ymin <= gcps.rel_y, gcps.rel_y < ymax])

        if np.count_nonzero(valid_pts) == 0:
            continue

        this_im_mes = E.MesureAppuiFlottant1Im(E.NameIm(im))

        for i, (ind, row) in enumerate(impts[valid_pts].iterrows()):
            this_mes = E.OneMesureAF1I(E.NamePt(gcps.iloc[ind]['id']),
                                       E.PtIm('{} {}'.format(row.j, row.i)))
            this_im_mes.append(this_mes)

        MesureSet.append(this_im_mes)

    tree = etree.ElementTree(MesureSet)
    tree.write(os.path.join(outdir, 'AutoMeasures{}-S2D.xml'.format(sub)),
               pretty_print=True, xml_declaration=True, encoding="utf-8")


def write_auto_gcps(gcp_df, sub, outdir, utm_zone, outname='AutoGCPs'):
    """
    Write GCP name, x, y, and z information to a text file to use with mm3d GCPConvert.

    :param pandas.DataFrame gcp_df: a DataFrame with the GCPs to save.
    :param str sub: the name of the block, if multiple blocks are being used (e.g., '_block1'). If not, use ''.
    :param str outdir: the output directory to save the files to.
    :param str utm_zone: the UTM zone name (e.g., 8N).
    :param str outname: the name to use for the GCPs file (default: AutoGCPs.txt)
    """
    with open(os.path.join(outdir, '{}{}.txt'.format(outname, sub)), 'w') as f:
        # print('#F= N X Y Z Ix Iy Iz', file=f)
        print('#F= N X Y Z', file=f)
        print('#Here the coordinates are in UTM {} X=Easting Y=Northing Z=Altitude'.format(utm_zone), file=f)
        for i, row in gcp_df.iterrows():
            # print('{} {} {} {} {} {} {}'.format(row.id, row.geometry.x, row.geometry.y, row.elevation,
            #                                        5/row.z_corr, 5/row.z_corr, 1), file=f)
            print('{} {} {} {}'.format(row.id, row.geometry.x, row.geometry.y, row.elevation), file=f)


def remove_measure(fn_meas, name):
    """
    Remove all instances of a given measure from an xml file.

    :param str fn_meas: the xml file (e.g., AutoMeasures-S2D.xml)
    :param str name: the measurement name (e.g., GCP0)
    """
    root = ET.parse(fn_meas).getroot()
    for im in root.findall('MesureAppuiFlottant1Im'):
        for mes in im.findall('OneMesureAF1I'):
            if mes.find('NamePt').text == name:
                im.remove(mes)

    tree = ET.ElementTree(root)
    tree.write(fn_meas, encoding="utf-8", xml_declaration=True)


def get_bascule_residuals(fn_basc, gcp_df):
    """
    Read a given GCPBascule residual file, and add the residuals to a DataFrame with GCP information.

    :param str fn_basc: the GCPBascule xml file to read the residuals from.
    :param pandas.DataFrame gcp_df: a DataFrame with the GCPs to read the residuals for.
    :return:
        - **gcp_df** (*pandas.DataFrame*) -- the input GCPs with the Bascule residuals added.
    """
    root = ET.parse(fn_basc).getroot()
    gcp_res = root.findall('Residus')
    gcp_names = np.array([res.find('Name').text for res in gcp_res])
    # residuals = np.array([float(res.find('Dist').text) for res in gcp_res])
    x_res = np.array([float(res.find('Offset').text.split()[0]) for res in gcp_res])
    y_res = np.array([float(res.find('Offset').text.split()[1]) for res in gcp_res])
    z_res = np.array([float(res.find('Offset').text.split()[2]) for res in gcp_res])
    dist = np.array([float(res.find('Dist').text) for res in gcp_res])

    for data_ in zip(gcp_names, x_res):
        gcp_df.loc[gcp_df.id == data_[0], 'xres'] = data_[1]

    for data_ in zip(gcp_names, y_res):
        gcp_df.loc[gcp_df.id == data_[0], 'yres'] = data_[1]

    for data_ in zip(gcp_names, z_res):
        gcp_df.loc[gcp_df.id == data_[0], 'zres'] = data_[1]

    for data_ in zip(gcp_names, dist):
        gcp_df.loc[gcp_df.id == data_[0], 'residual'] = data_[1]
    # gcp_df['residual'] = np.sqrt(gcp_df['xres'].values**2 + gcp_df['yres'].values**2)

    return gcp_df


def get_campari_residuals(fn_resids, gcp_df):
    """
    Read a given Campari residual file, and add the residuals to a DataFrame with GCP information.

    :param fn_resids: the Campari residual xml file to read.
    :param pandas.DataFrame gcp_df: a DataFrame with the GCPs to read the residuals for.
    :return:
        - **gcp_df** (*pandas.DataFrame*) -- the input GCPs with the Campari residuals added.
    """
    camp_root = ET.parse(fn_resids).getroot()

    last_iter = camp_root.findall('Iters')[-1].findall('OneAppui')
    camp_gcp_names = [a.find('Name').text for a in last_iter]

    err_max = []
    for a in last_iter:
        try:
            err_max.append(float(a.find('EcartImMax').text))
        except AttributeError:
            err_max.append(np.nan)

    camp_x = []
    camp_y = []
    camp_z = []

    for a in last_iter:
        try:
            camp_x.append(float(a.find('EcartFaiscTerrain').text.split()[0]))
            camp_y.append(float(a.find('EcartFaiscTerrain').text.split()[1]))
            camp_z.append(float(a.find('EcartFaiscTerrain').text.split()[2]))
        except AttributeError:
            camp_x.append(np.nan)
            camp_y.append(np.nan)
            camp_z.append(np.nan)

    for data_ in zip(camp_gcp_names, err_max):
        gcp_df.loc[gcp_df.id == data_[0], 'camp_res'] = data_[1]

    for data_ in zip(camp_gcp_names, camp_x):
        gcp_df.loc[gcp_df.id == data_[0], 'camp_xres'] = data_[1]

    for data_ in zip(camp_gcp_names, camp_y):
        gcp_df.loc[gcp_df.id == data_[0], 'camp_yres'] = data_[1]

    for data_ in zip(camp_gcp_names, camp_z):
        gcp_df.loc[gcp_df.id == data_[0], 'camp_zres'] = data_[1]

    return gcp_df


def move_bad_tapas(ori):
    """
    Read residual files output from Tapas (or Campari, GCPBascule), and move images with a NaN residual.

    :param str ori: the orientation directory to read the residuals file from (e.g., 'Ori-Relative').
    """
    root = ET.parse(os.path.join(ori, 'Residus.xml')).getroot()
    res_df = pd.DataFrame()

    nimgs = [len(a.findall('OneIm')) for a in root.findall('Iters')]

    nmax = max(nimgs)

    ind = np.where(np.array(nimgs) == nmax)[0].min()

    res_df['name'] = [a.find('Name').text for a in root.findall('Iters')[ind].findall('OneIm')]
    res_df['residual'] = [float(a.find('Residual').text) for a in root.findall('Iters')[ind].findall('OneIm')]
    res_df['pct_ok'] = [float(a.find('PercOk').text) for a in root.findall('Iters')[ind].findall('OneIm')]
    res_df['npts'] = [float(a.find('NbPts').text) for a in root.findall('Iters')[ind].findall('OneIm')]

    mkdir_p('bad')
    for im in res_df['name'][np.isnan(res_df.residual)]:
        print('{} -> bad/{}'.format(im, im))
        shutil.move(im, 'bad')


def run_bascule(in_gcps, outdir, img_pattern, sub, ori):
    """
    Interface for running mm3d GCPBascule and reading the residuals from the resulting xml file.

    :param pandas.DataFrame in_gcps: a DataFrame with the GCPs that are being input to Campari.
    :param str outdir: the output directory where the AutoGCPs.xml file is saved.
    :param str img_pattern: the match pattern for the images being input to Campari (e.g., "OIS.*tif")
    :param str sub: the name of the block, if multiple blocks are being used (e.g., '_block1'). If not, use ''.
    :param str ori: the name of the orientation directory (e.g., Ori-Relative).
    :return:
        - **out_gcps** (*pandas.DataFrame*) -- the input gcps with the updated Campari residuals.
    """
    echo = subprocess.Popen('echo', stdout=subprocess.PIPE)
    p = subprocess.Popen(['mm3d', 'GCPBascule', img_pattern, ori,
                          'TerrainRelAuto{}'.format(sub),
                          os.path.join(outdir, 'AutoGCPs{}.xml'.format(sub)),
                          os.path.join(outdir, 'AutoMeasures{}-S2D.xml'.format(sub))], stdin=echo.stdout)
    p.wait()

    out_gcps = get_bascule_residuals(os.path.join('Ori-TerrainRelAuto{}'.format(sub),
                                                  'Result-GCP-Bascule.xml'), in_gcps)
    return out_gcps


def run_campari(in_gcps, outdir, img_pattern, sub, dx, ortho_res, allfree=True):
    """
    Interface for running mm3d Campari and reading the residuals from the residual xml file.

    :param pandas.DataFrame in_gcps: a DataFrame with the GCPs that are being input to Campari.
    :param str outdir: the output directory where the AutoGCPs.xml file is saved.
    :param str img_pattern: the match pattern for the images being input to Campari (e.g., "OIS.*tif")
    :param str sub: the name of the block, if multiple blocks are being used (e.g., '_block1'). If not, use ''.
    :param int|float dx: the pixel resolution of the reference image.
    :param int|float ortho_res: the pixel resolution of the orthoimage being used.
    :param bool allfree: run Campari with AllFree=1 (True), or AllFree=0 (False). (default: True)
    :return:
        - **out_gcps** (*pandas.DataFrame*) -- the input gcps with the updated Campari residuals.
    """
    echo = subprocess.Popen('echo', stdout=subprocess.PIPE)
    if allfree:
        allfree_tag = 1
    else:
        allfree_tag = 0

    p = subprocess.Popen(['mm3d', 'Campari', img_pattern,
                          'TerrainRelAuto{}'.format(sub),
                          'TerrainFirstPass{}'.format(sub),
                          'GCP=[{},{},{},{}]'.format(os.path.join(outdir, 'AutoGCPs{}.xml'.format(sub)),
                                                     np.abs(dx),
                                                     os.path.join(outdir, 'AutoMeasures{}-S2D.xml'.format(sub)),
                                                     np.abs(dx / ortho_res)),
                          'SH=Homol', 'AllFree={}'.format(allfree_tag)], stdin=echo.stdout)
    p.wait()

    out_gcps = get_campari_residuals('Ori-TerrainFirstPass{}/Residus.xml'.format(sub), in_gcps)
    # out_gcps.dropna(inplace=True)  # sometimes, campari can return no information for a gcp
    return out_gcps


def save_gcps(in_gcps, outdir, utmstr, sub):
    """
    Save a GeoDataFrame of GCP information to shapefile, txt, and xml formats.

    After running, the following new files will be created:

        - outdir/AutoGCPs.shp (+ associated files)
        - outdir/AutoGCPs.txt
        - outdir/AutoGCPs.xml (output from mm3d GCPConvert)
        - outdir/AutoMeasures.xml (a file with image locations for each GCP)

    :param GeoDataFrame in_gcps: the gcps GeoDataFrame to save
    :param str outdir: the output directory to save the files to
    :param str utmstr: a UTM string generated by register.get_utm_str()
    :param str sub: the name of the block, if multiple blocks are being used (e.g., '_block1'). If not, use ''.

    """
    in_gcps.to_file(os.path.join(outdir, 'AutoGCPs{}.shp'.format(sub)))
    write_auto_gcps(in_gcps, sub, outdir, utmstr)
    echo = subprocess.Popen('echo', stdout=subprocess.PIPE)
    p = subprocess.Popen(['mm3d', 'GCPConvert', 'AppInFile',
                          os.path.join(outdir, 'AutoGCPs{}.txt'.format(sub))], stdin=echo.stdout)
    p.wait()

    auto_root = ET.parse(os.path.join(outdir, 'AutoMeasures{}-S2D.xml'.format(sub))).getroot()
    for im in auto_root.findall('MesureAppuiFlottant1Im'):
        for pt in im.findall('OneMesureAF1I'):
            if pt.find('NamePt').text not in in_gcps.id.values:
                im.remove(pt)

    # save AutoMeasures
    out_xml = ET.ElementTree(auto_root)
    out_xml.write(os.path.join(outdir, 'AutoMeasures{}-S2D.xml'.format(sub)),
                  encoding="utf-8", xml_declaration=True)


######################################################################################################################
# orientation tools - used for visualizing, manipulating camera orientation files
######################################################################################################################
def load_orientation(fn_img, ori):
    """
    Read camera position and rotation information from an Orientation xml file.

    :param str fn_img: the name of the image to read the orientation file for.
    :param str ori: the name of the orientation directory (e.g., Ori-Relative).
    :return:
        - **centre** (*list*) -- the camera position (x, y, z)
        - **l1** (*list*) -- the L1 orientation parameters
        - **l2** (*list*) -- the L2 orientation parameters
        - **l3** (*list*) -- the L3 orientation parameters
        - **prof** (*float*) -- the 'Profondeur' value from the xml file.
        - **altisol** (*float*) -- the 'AltiSol' value from the xml file.

    """
    ori_root = ET.parse(os.path.join(ori, 'Orientation-{}.xml'.format(fn_img))).getroot()
    if ori_root.tag != 'OrientationConique':
        ori_coniq = ori_root.find('OrientationConique')
    else:
        ori_coniq = ori_root
    centre = [float(p) for p in ori_coniq.find('Externe').find('Centre').text.split()]
    rotMat = ori_coniq.find('Externe').find('ParamRotation').find('CodageMatr')

    if rotMat is not None:
        l1 = [float(p) for p in rotMat.find('L1').text.split()]
        l2 = [float(p) for p in rotMat.find('L2').text.split()]
        l3 = [float(p) for p in rotMat.find('L3').text.split()]
    else:
        l1 = [np.nan, np.nan, np.nan]
        l2 = [np.nan, np.nan, np.nan]
        l3 = [np.nan, np.nan, np.nan]

    prof = ori_coniq.find('Externe').find('Profondeur')
    if prof is not None:
        prof = prof.text
    else:
        prof = np.nan

    altisol = ori_coniq.find('Externe').find('AltiSol')
    if altisol is not None:
        altisol = altisol.text
    else:
        altisol = np.nan

    return centre, l1, l2, l3, prof, altisol


def load_all_orientation(ori, imlist=None):
    """
    Load all of the orientation parameters for a set of images from a given directory.

    :param str ori: the orientation directory to read
    :param list imlist: the images to load. If not set, loads all orientation files from the given directory.
    :return:
        - **df** (*pandas.DataFrame*) -- a DataFrame containing the orientation parameters for each image
    """
    df = pd.DataFrame()
    points = []

    if imlist is None:
        imlist = [os.path.basename(g).split('Orientation-')[1].split('.xml')[0] for g in
                  glob(os.path.join(ori, '*.tif.xml'))]
        imlist.sort()

    for i, fn_img in enumerate(imlist):
        centre, l1, l2, l3, prof, altisol = load_orientation(fn_img, ori)

        df.loc[i, 'name'] = fn_img
        points.append(Point(centre[0], centre[1], centre[2]))
        df.loc[i, 'x'] = centre[0]
        df.loc[i, 'y'] = centre[1]
        df.loc[i, 'z'] = centre[2]

        df.loc[i, 'l11'] = l1[0]
        df.loc[i, 'l12'] = l1[1]
        df.loc[i, 'l13'] = l1[2]

        df.loc[i, 'l21'] = l2[0]
        df.loc[i, 'l22'] = l2[1]
        df.loc[i, 'l23'] = l2[2]

        df.loc[i, 'l31'] = l3[0]
        df.loc[i, 'l32'] = l3[1]
        df.loc[i, 'l33'] = l3[2]

        df.loc[i, 'profondeur'] = prof
        df.loc[i, 'altisol'] = altisol

    df['geometry'] = points
    return df


def extend_line(df, first, last):
    """
    Extend a flightline using existing camera positions.

    :param GeoDataFrame df: a GeoDataFrame containing the camera positions and image names
    :param str first: the name of the image to start interpolating from.
    :param str last: the name of the image to end interpolating at.
    :return:
        - **outpt** (*shapely.Point*) -- the new point along the flightline.
    """
    firstImg = df.loc[df.name.str.contains(first), 'geometry'].values[0]
    lastImg = df.loc[df.name.str.contains(last), 'geometry'].values[0]
    dx = firstImg.x - lastImg.x
    dy = firstImg.y - lastImg.y
    dz = firstImg.z - lastImg.z
    outpt = Point(firstImg.x + dx, firstImg.y + dy, firstImg.z + dz)

    return outpt


def interp_line(df, first, last, nimgs=None, pos=None):
    """
    Interpolate camera positions along a flightline.

    :param GeoDataFrame df: a GeoDataFrame containing the camera positions and image names
    :param str first: the name of the image to start interpolating from.
    :param str last: the name of the image to end interpolating at.
    :param int nimgs: the number of images to interpolate (default: calculated based on the image numbers)
    :param int pos: which image position to return (default: all images between first and last)
    :return:
        - **ptList** (*list*) -- a list containing the interpolated camera positions (or, a tuple of the requested position).
    """
    if nimgs is None:
        nimgs = np.abs(int(last) - int(first)).astype(int)
    firstImg = df.loc[df.name.str.contains(first), 'geometry'].values[0]
    lastImg = df.loc[df.name.str.contains(last), 'geometry'].values[0]
    flightLine = LineString([firstImg, lastImg])
    avgDist = flightLine.length / nimgs
    if pos is not None:
        return flightLine.interpolate(pos * avgDist)
    else:
        ptList = []
        for i in range(1, nimgs):
            ptList.append((i, flightLine.interpolate(i * avgDist)))
        return ptList


def update_center(fn_img, ori, new_center):
    """
    Update the camera position in an Orientation file.

    :param str fn_img: the name of the image to update the orientation for (e.g., 'OIS-Reech_ARCSEA000590122.tif')
    :param str ori: the name of the orientation directory (e.g., 'Ori-Relative')
    :param list new_center: a list of the new camera position [x, y, z]
    """
    ori_root = ET.parse(os.path.join(ori, 'Orientation-{}.xml'.format(fn_img))).getroot()
    if ori_root.tag != 'OrientationConique':
        ori_coniq = ori_root.find('OrientationConique')
    else:
        ori_coniq = ori_root

    ori_coniq.find('Externe').find('Centre').text = ' '.join([str(f) for f in new_center])

    tree = ET.ElementTree(ori_root)
    tree.write(os.path.join(ori, 'Orientation-{}.xml'.format(fn_img)),
               encoding="utf-8", xml_declaration=True)


def fix_orientation(cameras, ori_df, ori, nsig=4):
    """
    Correct erroneous Tapas camera positions using an estimated affine transformation between the absolute camera locations
    and the relative locations read from the orientation directory.

    Once the positions have been updated, you should re-run Tapas using the InOri set to the directory; e.g., if you
    have updated Ori-Relative, you should run:

        mm3d Tapas RadialBasic "OIS.*tif" InOri=Relative Out=Relative LibFoc=0

    :param pandas.DataFrame cameras: A DataFrame containing camera positions (x, y, z) and a 'name' column that contains
        the image names.
    :param pandas.DataFrame ori_df: A DataFrame output from sPyMicMac.micmac.load_all_orientations, or that contains
        a 'name' column and camera positions in relative space (x, y, z)
    :param str ori: the Orientation directory to update (e.g., Ori-Relative)
    :param int|float nsig: the number of normalized absolute deviations from the median residual value to consider
        a camera an outlier (default: 4)
    """
    join = cameras.set_index('name').join(ori_df.set_index('name'), lsuffix='abs', rsuffix='rel')

    model = AffineTransform()
    model.estimate(join[['xabs', 'yabs']].values, join[['xrel', 'yrel']].values)

    res = model.residuals(join[['xabs', 'yabs']].values, join[['xrel', 'yrel']].values)

    outliers = res - np.median(res) > nsig * nmad(res)
    if np.count_nonzero(outliers) > 0:
        interp = LinearNDInterpolator(join.loc[~outliers, ['xrel', 'yrel']].values, join.loc[~outliers, 'zrel'])
        print('found {} outliers using nsig={}'.format(np.count_nonzero(outliers), nsig))
        for name, row in join[outliers].iterrows():
            new_x, new_y = model(row[['xabs', 'yabs']].values)[0]
            new_z = interp(new_x, new_y)

            print('new location for {}: {}, {}, {}'.format(name, new_x, new_y, new_z))
            print('writing new Orientation file for {}'.format(name))
            update_center(name, ori, [new_x, new_y, new_z])


######################################################################################################################
#
######################################################################################################################
def dem_to_text(fn_dem, fn_out='dem_pts.txt', spacing=100):
    """
    Write elevations from a DEM raster to a text file for use in mm3d PostProc Banana.

    :param str fn_dem: the filename of the DEM to read.
    :param str fn_out: the name of the text file to write out (default: dem_pts.txt)
    :param int spacing: the pixel spacing of the DEM to write (default: every 100 pixels)
    """
    if isinstance(fn_dem, GeoImg):
        dem = fn_dem
    else:
        dem = GeoImg(fn_dem)
    x, y = dem.xy()

    z = dem.img[::spacing, ::spacing].flatten()
    x = x[::spacing, ::spacing].flatten()
    y = y[::spacing, ::spacing].flatten()

    x = x[np.isfinite(z)]
    y = y[np.isfinite(z)]
    z = z[np.isfinite(z)]

    with open(fn_out, 'w') as f:
        for pt in list(zip(x, y, z)):
            print(pt[0], pt[1], pt[2], file=f)




"""
spymicmac.ee_tools is a collection of tools for getting GEE data for processing KH-9 Hexagon imagery.
"""
import ee


ee.Initialize()


def get_landsat_collections(filt_geom=None, sensors='all', tier='all'):
    """
    Return a collection of raw Landsat images filtered by an (optional) geometry, given a list of sensor names and
    tier levels. Default behavior is to select all sensors and all tiers.

    Valid sensor names: LC08, LE07, LT05, LT04, LM05, LM04, LM03, LM02, LM01

    Valid tier levels: T1, T2. Check Landsat docs for details.

    :param filt_geom: A geometry object to filter images by. See EE docs for more details.
    :param sensors: Landsat sensor(s) to select images from.
    :param tier: Processing tier(s) to select.
    :return:
    """
    assert tier.lower() in ['all', 't1', 't2'], "tier is not one of 'all', 't1', 't2': {}".format(tier)
    if tier == 'all':
        tiers = ['T1', 'T2']
    else:
        tiers = list(tier.upper())

    allSensors = ['LC08', 'LE07', 'LT05', 'LT04', 'LM05', 'LM04', 'LM03', 'LM02', 'LM01']
    if isinstance(sensors, str) and sensors.lower() == 'all':
        getSensors = allSensors
    else:
        getSensors = []
        if isinstance(sensors, list):
            for sens in sensors:
                assert sens.upper() in allSensors, "{} is not a recognized Landsat sensor name".format(sens)
                getSensors.append(sens)
        else:
            assert isinstance(sensors, str), "sensors is not a list of strings, or a single string: {}".format(sensors)
            assert sensors.upper() in allSensors, "{} is not a recognized Landsat sensor name".format(sensors)
            getSensors = [sensors]

    outColls = []
    for sens in getSensors:
        for t in tiers:
            if filt_geom is not None:
                outColls.append(ee.ImageCollection('LANDSAT/{}/C01/{}'.format(sens, t)).filterBounds(filt_geom))
            else:
                outColls.append(ee.ImageCollection('LANDSAT/{}/C01/{}'.format(sens, t)))

    if len(outColls) == 1:
        return outColls[0]
    else:
        _out = outColls[0]
        for coll in outColls[1:]:
            _out.merge(coll)
        return _out


def get_srtm(ver='usgs', filt_geom=None):
    '''
    Return the SRTM DEM, clipped to an optional geometry.

    :param ver: SRTM version to select, one of 'usgs', 'cgiar'
    :param filt_geom: A geometry object to clip images to

    :return:
    '''
    assert ver.lower() in ['usgs', 'cgiar'], "srtm version is not one of 'usgs', 'cgiar': {}".format(ver)

    verDict = {'usgs': 'USGS/SRTMGL1_003', 'cgiar': 'CGIAR/SRTM90_V4'}
    _ver = verDict[ver]

    if filt_geom is not None:
        return ee.Image(_ver).clip(filt_geom)
    else:
        return ee.Image(_ver)


def get_alos_dem(filt_geom=None):
    '''
    Return the ALOS DEM, clipped to an optional geometry.

    :param filt_geom: A geometry object to clip images to

    :return:
    '''

    if filt_geom is not None:
        return ee.Image('JAXA/ALOS/AW3D30/V2_2').select('AVE_DSM').clip(filt_geom)
    else:
        return ee.Image('JAXA/ALOS/AW3D30/V2_2').select('AVE_DSM')



def get_s2_collection():
    pass


def get_cloudfree_mosaic():
    pass



#!/usr/bin/env python
import argparse
from spymicmac.micmac import remove_measure


def _argparser():
    parser = argparse.ArgumentParser(description="Remove GCP(s) from a Measures xml file.",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('fn_meas', action='store', type=str, help='xml file to remove GCP(s) from.')
    parser.add_argument('gcp', action='store', type=str, nargs='+', help='GCP name(s) to remove from <fn_meas>.')
    return parser


def main():
    parser = _argparser()
    args = parser.parse_args()

    for gcp in args.gcp:

        remove_measure(args.fn_meas, gcp)


if __name__ == "__main__":
    main()



"""
sPyMicMac is a python package to help in the processing of KH-9 and historical aerial imagery using MicMac
"""
from . import image
from . import micmac
from . import register
from . import usgs
# from . import ee_tools



#!/usr/bin/env python
import argparse
from spymicmac.image import remove_crosses


def _argparser():
    parser = argparse.ArgumentParser(description="Remove Reseau marks from KH-9 image(s).",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('img', action='store', type=str, nargs='+', help='Image(s) to remove crosses from.')
    return parser


def main():
    parser = _argparser()
    args = parser.parse_args()

    for im in args.img:
        print(im)
        remove_crosses(im)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import argparse
from glob import glob
from spymicmac.image import join_hexagon


def _argparser():
    parser = argparse.ArgumentParser(description="Join two halves of a scanned KH-9 Hexagon image (or four parts of a scanned KH-4 Corona image).",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-o', '--overlap', action='store', type=int, default=2000,
                        help='overlap search width between two images [2000]')
    parser.add_argument('-p', '--pattern', action='store', type=str, default='DZB',
                       help='Match pattern for images [DZB]')
    parser.add_argument('-b', '--blend', action='store_true',
                        help='Blend across image halves to prevent a sharp line at edge.')
    parser.add_argument('-c', '--corona', action='store_true',
                        help='Images are Corona KH-4/KH-4A scans (i.e., there are 4 parts).')
    return parser


def main():
    parser = _argparser()
    args = parser.parse_args()

    imlist = [f.split('_a.tif')[0] for f in glob('{}*a.tif'.format(args.pattern))]
    for im in imlist:
        print(im)
        join_hexagon(im,
                     overlap=args.overlap,
                     blend=args.blend,
                     corona=args.corona)


if __name__ == "__main__":
    main()



import os
import argparse
from glob import glob
from skimage.io import imread, imsave
from pybob.bob_tools import mkdir_p
from spymicmac.image import balance_image


def _argparser():
    _parser = argparse.ArgumentParser(description="Apply Contrast-limited Adaptive Histogram Equalization (CLAHE) to all re-sampled images in current directory.",
                                      formatter_class=argparse.RawDescriptionHelpFormatter)
    return _parser


def main():
    parser = _argparser()
    args = parser.parse_args()

    mkdir_p('balanced')

    imlist = glob('OIS*.tif')
    imlist.sort()

    for im in imlist:
        print(im)
        img_filt = balance_image(imread(im))
        imsave(os.path.join('balanced', im), img_filt)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import argparse
from spymicmac.micmac import get_gcp_meas
import lxml.etree as etree
import lxml.builder as builder


def _argparser():
    parser = argparse.ArgumentParser(description="Combine outputs of XYZ2Im into single xml file for further processing",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('ij_files', action='store', type=str, nargs='+', help='txt files containing image i,j points')
    parser.add_argument('-o', '--out_file', action='store', type=str, default='MeasuresAuto.xml',
                        help='output filename [AutoMeasures.xml]')
    parser.add_argument('-n', '--no_distortion', action='store_true',
                        help='Use gcp locations computed assuming no distortion to filter GCPs from images.')
    return parser


def main():
    parser = _argparser()
    args = parser.parse_args()

    E = builder.ElementMaker()
    MesureSet = E.SetOfMesureAppuisFlottants()
    for ij_file in args.ij_files:
        imname = ij_file.split('Auto-')[1].split('.txt')[0]
        if args.no_distortion:
            nodist_file = ij_file.replace('Auto', 'NoDist')
        else:
            nodist_file = None

        print(imname)
        this_meas = get_gcp_meas(imname, ij_file, '.', E, nodist=nodist_file)
        MesureSet.append(this_meas)

    tree = etree.ElementTree(MesureSet)
    tree.write(args.out_file, pretty_print=True, xml_declaration=True, encoding="utf-8")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import argparse
from spymicmac.register import register_ortho


def _argparser():
    _parser = argparse.ArgumentParser(description="Register a relative orthoimage and DEM to a reference orthorectified image and DEM.",
                                      formatter_class=argparse.RawDescriptionHelpFormatter)

    _parser.add_argument('fn_ortho', action='store', type=str, help='non-referenced orthophoto mosaic')
    _parser.add_argument('fn_ref', action='store', type=str, help='georeferenced satellite image')
    _parser.add_argument('fn_reldem', action='store', type=str, help='relative dem corresponding to ortho')
    _parser.add_argument('fn_dem', action='store', type=str, help='dem')
    _parser.add_argument('-glacmask', action='store', type=str, default=None,
                         help='path to shapefile of glacier outlines (i.e., an exclusion mask)')
    _parser.add_argument('-landmask', action='store', type=str, default=None,
                         help='path to shapefile of land outlines (i.e., an inclusion mask)')
    _parser.add_argument('-footprints', action='store', type=str, default=None,
                         help='path to shapefile of image outlines. If not set, will download from USGS.')
    _parser.add_argument('-im_subset', action='store', type=str, default=None, nargs='+',
                         help='subset of raw images to work with (default all)')
    _parser.add_argument('-b', '--block_num', action='store', type=str, default=None,
                         help='Block number to use if multiple image blocks exist in directory.')
    _parser.add_argument('-ori', action='store', type=str, default='Relative',
                         help='name of orientation directory (after Ori-) [Relative]')
    _parser.add_argument('-ortho_res', action='store', type=float, default=8,
                         help='approx. ground sampling distance (pixel resolution) of ortho image. [8 m]')
    _parser.add_argument('-imgsource', action='store', type=str, default='DECLASSII',
                         help='USGS dataset name for images [DECLASSII]')
    _parser.add_argument('-density', action='store', type=int, default=200,
                         help='pixel spacing to look for GCPs [200]')
    _parser.add_argument('-no_allfree', action='store_false',
                         help='run Campari with AllFree set to False')
    return _parser


def main():
    parser = _argparser()
    args = parser.parse_args()

    register_ortho(args.fn_ortho, args.fn_ref, args.fn_reldem, args.fn_dem,
                   glacmask=args.glacmask,
                   landmask=args.landmask,
                   footprints=args.footprints,
                   im_subset=args.im_subset,
                   block_num=args.block_num,
                   ori=args.ori,
                   ortho_res=args.ortho_res,
                   imgsource=args.imgsource,
                   density=args.density,
                   allfree=args.no_allfree)


if __name__ == "__main__":
    main()



"""
spymicmac.register is a collection of tools for registering images and finding GCPs.
"""
import os
import re
import subprocess
import shutil
import numpy as np
import matplotlib.pyplot as plt
from osgeo import gdal
import pandas as pd
import geopandas as gpd
from PIL import Image
from glob import glob
from shapely.geometry.point import Point
from skimage.io import imread
from skimage.measure import ransac
from skimage.filters import median
from skimage.morphology import disk, binary_dilation
from skimage.transform import EuclideanTransform, AffineTransform, warp
from pybob.bob_tools import mkdir_p
from pybob.ddem_tools import nmad
from pybob.image_tools import create_mask_from_shapefile
from pybob.GeoImg import GeoImg
import spymicmac.image as imtools
import spymicmac.micmac as mmtools
from spymicmac.usgs import get_usgs_footprints


def sliding_window_filter(img_shape, pts_df, winsize, stepsize=None, mindist=2000, how='residual', is_ascending=True):
    """
    Given a DataFrame of indices representing points, use a sliding window filter to keep only the 'best' points within
    a given window and separation distance.

    :param array-like img_shape: The shape of the image to filter indices from.
    :param DataFrame pts_df: A DataFrame that contains the pixel locations of points (column: orig_j, row: orig_i)
    :param int winsize: the size of the window to use for the filter.
    :param int stepsize: how large of a step size to use for the sliding window (default: winsize/2)
    :param int mindist: the minimum distance (in pixels) to use between points (default: 2000)
    :param str how: how to sort pts_df to determine the 'best' points to keep (default: 'residual')
    :param bool is_ascending: whether the column named in 'how' should be sorted ascending or descending (default: True)
    :return:
        - **out_inds** (*array-like*) -- an array with the filtered indices
    """
    if stepsize is None:
        stepsize = winsize / 2

    _out_inds = []
    _out_pts = []

    for x_ind in np.arange(stepsize, img_shape[1], winsize):
        for y_ind in np.arange(stepsize, img_shape[0], winsize):
            min_x = x_ind - winsize / 2
            max_x = x_ind + winsize / 2
            min_y = y_ind - winsize / 2
            max_y = y_ind + winsize / 2
            samp_ = pts_df.loc[np.logical_and.reduce([pts_df.orig_i > min_x,
                                                      pts_df.orig_i < max_x,
                                                      pts_df.orig_j > min_y,
                                                      pts_df.orig_j < max_y])].copy()
            if samp_.shape[0] == 0:
                continue
            # only take the above-average z_corr values
            samp_ = samp_[samp_.z_corr >= samp_.z_corr.quantile(0.5)]
            # make sure we get the best residual
            samp_.sort_values(how, ascending=is_ascending, inplace=True)
            if len(_out_inds) == 0:
                best_ind = samp_.index[0]
                best_pt = Point(samp_.loc[best_ind, ['orig_j', 'orig_i']].values)

                _out_inds.append(best_ind)
                _out_pts.append(best_pt)
            else:
                for _ind, _row in samp_.iterrows():
                    this_pt = Point(_row[['orig_j', 'orig_i']].values)
                    this_min_dist = np.array([this_pt.distance(pt) for pt in _out_pts]).min()
                    if this_min_dist > mindist:
                        _out_inds.append(_ind)
                        _out_pts.append(this_pt)

    return np.array(_out_inds)


def get_imlist(im_subset):
    """
    Given either a list of filenames or a regex match pattern, return a list of filenames and a pattern to provide
    to MicMac.

    :param list im_subset: a list of filenames or a match pattern (e.g., ['OIS.*tif']) representing filenames
    :return:
        - **imlist** (*list*) -- the list of filenames matching the provided pattern.
        - **match_pattern** (*str*) -- the match pattern to be provided to MicMac.
    """
    if im_subset is None:
        imlist = glob('OIS*.tif')
        match_pattern = 'OIS.*.tif'
    else:
        if len(im_subset) > 1:
            imlist = im_subset
            # match_pattern = mmtools.get_match_pattern(imlist)
            match_pattern = '|'.join(imlist)
        else:
            match_pattern = im_subset[0] + '.*tif'
            imlist = [f for f in glob('OIS*.tif') if re.search(match_pattern, f)]

    return imlist, match_pattern


def get_lowres_transform(lowres, ref, fprint, ref_mask, imgsource='DECLASSII'):

    if imgsource == 'DECLASSII':
        M = imtools.transform_from_fprint(lowres, ref, fprint, ref_mask)
    else:
        try:
            ortho_mask = 255 * np.ones(lowres.shape, dtype=np.uint8)
            ortho_mask[binary_dilation(lowres == 0, selem=disk(5))] = 0

            kp, des, matches = imtools.get_matches(lowres.astype(np.uint8),
                                                   ref.img.astype(np.uint8),
                                                   mask1=ortho_mask,
                                                   mask2=ref_mask,
                                                   dense=True)
            print('{} matches found.'.format(len(matches)))

            src_pts = np.array([kp[0][m.queryIdx].pt for m in matches])
            dst_pts = np.array([kp[1][m.trainIdx].pt for m in matches])

            M, inliers = ransac((dst_pts, src_pts), EuclideanTransform,
                                min_samples=10, residual_threshold=20, max_trials=10000)
            resids = M.residuals(dst_pts, src_pts)[inliers]
            print('{} matches used for initial transformation'.format(np.count_nonzero(inliers)))

        except ValueError:
            raise RuntimeError("Unable to find initial transformation with the parameters provided.")

    return M


def get_utm_str(epsg):
    """
    Given a GeoImg, read the associated EPSG code and return a UTM zone.

    Examples:
        - get_utm_str(32608) -> 'UTM Zone 8N'
        - get_utm_str(32708) -> 'UTM Zone 8S'

    :param int|str epsg: a str or int representing an EPSG Code
    :return:
        - **utm_str** (*str*) -- the UTM string representation
    """
    epsg_str = str(epsg)
    hemi_dict = {'6': 'N', '7': 'S'}

    if epsg_str[:2] == '32':
        utm_str = epsg_str[-2:] + hemi_dict[epsg_str[2]]
    else:
        utm_str = 'not utm'

    return utm_str


def transform_centers(img_gt, ref, imlist, footprints, ori):
    """
    Use the camera centers in relative space provided by MicMac Orientation files, along with camera footprints,
    to estimate a transformation between the relative coordinate system and the absolute coordinate system.

    :param array-like img_gt: the image GeoTransform (as read from a TFW file)
    :param GeoImg ref: the reference image to use to determine the output image shape
    :param list imlist: a list of of the images that were used for the relative orthophoto
    :param GeoDataFrame footprints: the (approximate) image footprints - the centroid will be used for the absolute camera positions.
    :param str ori: name of orientation directory
    :return:
        - **model** (*AffineTransform*) -- the estimated Affine Transformation between relative and absolute space
        - **inliers** (*array-like*) -- a list of the inliers returned by skimage.measure.ransac
        - **join** (*GeoDataFrame*) -- the joined image footprints and relative orientation files
    """

    rel_ori = mmtools.load_all_orientation(ori, imlist=imlist)

    footprints = footprints.to_crs(epsg=ref.epsg).copy()

    for i, row in footprints.iterrows():
        footprints.loc[i, 'x'] = row['geometry'].centroid.x
        footprints.loc[i, 'y'] = row['geometry'].centroid.y
        footprints.loc[i, 'name'] = 'OIS-Reech_' + row['ID'] + '.tif'

    join = footprints.set_index('name').join(rel_ori.set_index('name'), lsuffix='abs', rsuffix='rel')
    join.dropna(inplace=True)

    ref_ij = np.array([ref.xy2ij((row.xabs, row.yabs)) for i, row in join.iterrows()])
    rel_ij = np.array([((row.xrel - img_gt[4]) / img_gt[0],
                        (row.yrel - img_gt[5]) / img_gt[3]) for i, row in join.iterrows()])

    model, inliers = ransac((ref_ij[:, ::-1], rel_ij), AffineTransform, min_samples=10, residual_threshold=100, max_trials=5000)
    print('{} inliers for center transformation'.format(np.count_nonzero(inliers)))
    return model, inliers, join


def refine_lowres_tfm(rough_tfm, ref, mask, Minit):
    rough_gcps = imtools.find_grid_matches(rough_tfm, ref, mask, Minit, spacing=20, srcwin=40, dstwin=100)
    max_d = 100 - 40

    rough_gcps = rough_gcps[np.logical_and.reduce([rough_gcps.di.abs() < max_d,
                                                   rough_gcps.di < rough_gcps.di.median() + 2 * nmad(rough_gcps.di),
                                                   rough_gcps.di > rough_gcps.di.median() - 2 * nmad(rough_gcps.di),
                                                   rough_gcps.dj.abs() < max_d,
                                                   rough_gcps.dj < rough_gcps.dj.median() + 2 * nmad(rough_gcps.dj),
                                                   rough_gcps.dj > rough_gcps.dj.median() - 2 * nmad(rough_gcps.dj)])]

    Mref, inliers = ransac((rough_gcps[['search_j', 'search_i']].values,
                            rough_gcps[['orig_j', 'orig_i']].values),
                           AffineTransform, min_samples=10, residual_threshold=10, max_trials=5000)
    return Mref, inliers


def scale_factor(centers, img_gt, ref):

    ref_ij = np.array([ref.xy2ij((row.xabs, row.yabs)) for i, row in centers.iterrows()])
    rel_ij = np.array([((row.xrel - img_gt[4]) / img_gt[0],
                        (row.yrel - img_gt[5]) / img_gt[3]) for i, row in centers.iterrows()])
    ref_pt = Point(ref_ij[0])
    rel_pt = Point(rel_ij[0])

    ref_dist = []
    rel_dist = []

    for i in range(1, ref_ij.shape[0]):
        ref_dist.append(ref_pt.distance(Point(ref_ij[i])))
        rel_dist.append(rel_pt.distance(Point(rel_ij[i])))

    return np.nanmean(np.array(ref_dist) / np.array(rel_dist))


def get_mask(footprints, img, imlist, landmask=None, glacmask=None):
    """
    Create a mask for an image from different sources.

    :param GeoDataFrame footprints: vector data representing image footprints
    :param GeoImg img: the GeoImg to create a mask for
    :param array-like imlist: a list of image names
    :param str landmask: path to file of land outlines (i.e., an inclusion mask)
    :param str glacmask: path to file of glacier outlines (i.e., an exclusion mask)

    :returns:
        - **mask** (*array-like*) -- the mask
        - **fmask** (*GeoImg*) -- the georeferenced footprint mask
        - **img** (*GeoImg*) -- the GeoImg, cropped to a 10 pixel buffer around the image footprints
    """
    fmask, fprint = imtools.get_footprint_mask(footprints, img, imlist, fprint_out=True)
    fmask_geo = img.copy(new_raster=fmask)

    xmin, ymin, xmax, ymax = fprint.buffer(img.dx * 10).bounds

    img = img.crop_to_extent([xmin, xmax, ymin, ymax], pixel_size=img.dx)

    fmask = fmask_geo.crop_to_extent([xmin, xmax, ymin, ymax], pixel_size=img.dx).img == 1

    mask = imtools.make_binary_mask(img.img, erode=3, mask_value=np.nan)
    if landmask is not None:
        lmask = create_mask_from_shapefile(img, landmask)
        mask[~lmask] = 0
    if glacmask is not None:
        gmask = create_mask_from_shapefile(img, glacmask)
        mask[gmask] = 0

    mask[~fmask] = 0

    return mask, fmask, img


def register_ortho_old(fn_ortho, fn_ref, fn_reldem, fn_dem, glacmask=None, landmask=None, footprints=None,
                   im_subset=None, tfm_points=None, block_num=None, ori='Relative',
                   init_res=400, ortho_res=8, imgsource='DECLASSII', density=200, out_dir=None):
    
    print('start.')

    if out_dir is None:
        out_dir = 'auto_gcps'

    mkdir_p(out_dir)

    if block_num is not None:
        subscript = '_block{}'.format(block_num)
    else:
        subscript = ''

    ort_dir = os.path.dirname(fn_ortho)

    imlist, match_pattern = get_imlist(im_subset)

    ref_img = GeoImg(fn_ref)

    utm_str = get_utm_str(ref_img.epsg)

    ortho = imread(fn_ortho)

    ortho = median(ortho, selem=disk(1))

    ortho_ = Image.fromarray(ortho)

    lowres_ = int(init_res / ortho_res)

    ortho_lowres = np.array(ortho_.resize((np.array(ortho_.size) / lowres_).astype(int), Image.LANCZOS))

    ref_lowres = ref_img.resample(init_res)

    if footprints is not None:
        fmask, fprint = imtools.get_footprint_mask(footprints, ref_lowres, imlist, fprint_out=True)
    else:
        clean_imlist = [im.split('OIS-Reech_')[-1].split('.tif')[0] for im in imlist]
        print('Attempting to get image footprints from USGS EarthExplorer.')
        _fprints = get_usgs_footprints(clean_imlist, dataset=imgsource)
        fmask, fprint = imtools.get_footprint_mask(_fprints, ref_lowres, imlist, fprint_out=True)

    fmask_geo = ref_lowres.copy(new_raster=fmask)

    xmin, ymin, xmax, ymax = fprint.buffer(init_res * 10).bounds

    ref_img = ref_img.crop_to_extent([xmin, xmax, ymin, ymax], pixel_size=ref_img.dx)

    if isinstance(ref_lowres.img, np.ma.masked_array):
        ref_eq = imtools.stretch_image(ref_lowres.img.data.copy(), scale=(0.02, 0.98), mask=fmask)
    else:
        ref_eq = imtools.stretch_image(ref_lowres.img.copy(), scale=(0.02, 0.98), mask=fmask)    

    ref_lowres = ref_lowres.copy(new_raster=ref_eq, datatype=gdal.GDT_Byte)

    ref_lowres = ref_lowres.crop_to_extent([xmin, xmax, ymin, ymax], pixel_size=init_res)
    fmask = fmask_geo.crop_to_extent([xmin, xmax, ymin, ymax], pixel_size=init_res).img == 1

    lowres_mask = imtools.make_binary_mask(ref_lowres.img, erode=3, mask_value=np.nan)
    if landmask is not None:
        lmask = create_mask_from_shapefile(ref_lowres, landmask)
        lowres_mask[~lmask] = 0
    if glacmask is not None:
        gmask = create_mask_from_shapefile(ref_lowres, glacmask)
        lowres_mask[gmask] = 0

    lowres_mask[~fmask] = 0

    if tfm_points is not None:
        pts = pd.read_csv(tfm_points)
        src = pts[['J', 'I']].values / lowres_
        dst = np.array([ref_lowres.xy2ij((row.X, row.Y)) for i, row in pts.iterrows()])
        M = EuclideanTransform()
        M.estimate(src, dst[:, ::-1])
    else:
        M = get_lowres_transform(ortho_lowres, ref_lowres, fprint, lowres_mask, imgsource=imgsource)

    init_tfm = warp(ortho_lowres, M, output_shape=ref_lowres.img.shape, preserve_range=True)

    # if glacmask is not None:
    #   lowres_mask[gmask] = 255
    #    lowres_mask[~fmask] = 0

    rough_gcps = imtools.find_grid_matches(init_tfm, ref_lowres, lowres_mask, M, spacing=20, srcwin=40, dstwin=100)
    max_d = 100 - 40

    rough_gcps = rough_gcps[fmask[rough_gcps.search_i, rough_gcps.search_j] > 0]

    rough_gcps = rough_gcps[np.logical_and.reduce([rough_gcps.di.abs() < max_d,
                                                   rough_gcps.di < rough_gcps.di.median() + 2 * nmad(rough_gcps.di),
                                                   rough_gcps.di > rough_gcps.di.median() - 2 * nmad(rough_gcps.di),
                                                   rough_gcps.dj.abs() < max_d,
                                                   rough_gcps.dj < rough_gcps.dj.median() + 2 * nmad(rough_gcps.dj),
                                                   rough_gcps.dj > rough_gcps.dj.median() - 2 * nmad(rough_gcps.dj)])]

    Minit, inliers = ransac((rough_gcps[['search_j', 'search_i']].values,
                             rough_gcps[['orig_j', 'orig_i']].values),
                            AffineTransform, min_samples=10, residual_threshold=10, max_trials=5000)

    rough_gcps['residuals'] = Minit.residuals(rough_gcps[['search_j', 'search_i']].values,
                                              rough_gcps[['orig_j', 'orig_i']].values)
    # inliers = rough_gcps.residuals < min(10 * nmad(rough_gcps.residuals), 25)
    print('{} valid matches used for initial transformation'.format(np.count_nonzero(inliers)))

    ortho_tfm = warp(ortho_lowres, Minit, output_shape=ref_lowres.img.shape, preserve_range=True, order=5)

    dst_scale = ref_lowres.dx / ref_img.dx

    fig, ax = plt.subplots(1, 2, figsize=(7, 5))
    ax[0].imshow(ortho_tfm, cmap='gray')
    ax[1].imshow(ref_lowres.img, cmap='gray')

    plt.savefig('initial_transformation{}.png'.format(subscript), dpi=200, bbox_inches='tight')
    plt.close(fig)

    i_ = np.arange(0, ortho_lowres.shape[0], 10)
    j_ = np.arange(0, ortho_lowres.shape[1], 10)

    I, J = np.meshgrid(i_, j_)

    src_grd = np.array(list(zip(J.reshape(-1, 1), I.reshape(-1, 1)))).reshape(-1, 2)

    dst_tfm = []
    for pt in src_grd:
        dst_tfm.append(Minit.inverse(pt) * dst_scale)
    dst_tfm = np.array(dst_tfm).reshape(-1, 2)

    Minit_full, _ = ransac((dst_tfm, lowres_ * src_grd), AffineTransform, min_samples=3,
                           residual_threshold=1, max_trials=1000)
    # Minit_full = AffineTransform()
    # Minit_full.estimate((dst_tfm, lowres_ * src_grd))

    rough_tfm = warp(ortho, Minit_full, output_shape=ref_img.img.shape, preserve_range=True)

    mask_full = imtools.make_binary_mask(ref_img.img, mask_value=np.nan)
    if landmask is not None:
        lm = create_mask_from_shapefile(ref_img, landmask)
        mask_full[~lm] = 0
    if glacmask is not None:
        gm = create_mask_from_shapefile(ref_img, glacmask)
        mask_full[gm] = 0
    mask_full[rough_tfm == 0] = 0

    # for each of these pairs (src, dst), find the precise subpixel match (or not...)
    gcps = imtools.find_grid_matches(rough_tfm, ref_img, mask_full, Minit_full, spacing=density, dstwin=600)

    xy = np.array([ref_img.ij2xy((pt[1], pt[0])) for pt in gcps[['search_j', 'search_i']].values]).reshape(-1, 2)
    gcps['geometry'] = [Point(pt) for pt in xy]

    gcps = gcps[mask_full[gcps.search_i, gcps.search_j] == 255]

    max_d = 400 - 60

    # gcps = gcps[np.logical_and.reduce([gcps.di.abs() < max_d,
    #                                    gcps.di.abs() < 2 * nmad(gcps.di),
    #                                    gcps.dj.abs() < max_d,
    #                                    gcps.dj.abs() < 2 * nmad(gcps.dj)])]

    # gcps = gcps[gcps.z_corr > gcps.z_corr.quantile(0.5)]

    if os.path.exists(fn_ortho.replace('.tif', '.tfw')):
        fn_tfw = fn_ortho.replace('.tif', '.tfw')
        with open(fn_tfw, 'r') as f:
            gt = [float(l.strip()) for l in f.readlines()]

        gcps['rel_x'] = gt[4] + gcps['orig_j'].values * gt[0]  # need the original image coordinates
        gcps['rel_y'] = gt[5] + gcps['orig_i'].values * gt[3]

    gcps['elevation'] = 0
    gcps.crs = ref_img.proj4

    gcps.dropna(inplace=True)

    print('loading dems')
    dem = GeoImg(fn_dem, dtype=np.float32)
    rel_dem = GeoImg(fn_reldem)

    # gcps.crs = {'init': 'epsg:{}'.format(ref_img.epsg)}
    for i, row in gcps.to_crs(crs=dem.proj4).iterrows():
        gcps.loc[i, 'elevation'] = dem.raster_points([(row.geometry.x, row.geometry.y)], nsize=3, mode='linear')
        if os.path.exists(fn_ortho.replace('.tif', '.tfw')):
            gcps.loc[i, 'el_rel'] = rel_dem.raster_points([(row.rel_x, row.rel_y)], nsize=3, mode='linear')

    # drop any gcps where we don't have a DEM value or a valid match
    gcps.dropna(inplace=True)

    # run ransac to find the matches between the transformed image and the master image make a coherent transformation
    # residual_threshold is 10 pixels to allow for some local distortions, but get rid of the big blunders
    Mref, inliers_ref = ransac((gcps[['search_j', 'search_i']].values, gcps[['match_j', 'match_i']].values),
                               AffineTransform, min_samples=6, residual_threshold=20, max_trials=5000)

    gcps['aff_resid'] = Mref.residuals(gcps[['search_j', 'search_i']].values,
                                       gcps[['match_j', 'match_i']].values)

    gcps_orig = gcps.copy()

    # gcps = gcps[inliers_ref]

    out = sliding_window_filter([ortho.shape[1], ortho.shape[0]], gcps,
                                min(1000, ortho.shape[1] / 4, ortho.shape[0] / 4),
                                mindist=500, how='aff_resid', is_ascending=False)
    gcps = gcps.loc[out]

    # Mfin, inliers_fin = ransac((gcps[['search_j', 'search_i']].values, gcps[['match_j', 'match_i']].values),
    #                            AffineTransform, min_samples=6, residual_threshold=20, max_trials=5000)
    # gcps = gcps[inliers_fin]

    print('{} valid matches found'.format(gcps.shape[0]))

    gcps.index = range(gcps.shape[0])  # make sure index corresponds to row we're writing out
    gcps['id'] = ['GCP{}'.format(i) for i in range(gcps.shape[0])]
    gcps.to_file(os.path.join(out_dir, 'AutoGCPs{}.shp'.format(subscript)))

    print('writing AutoGCPs.txt')
    mmtools.write_auto_gcps(gcps, subscript, out_dir, utm_str)

    print('converting AutoGCPs.txt to AutoGCPs.xml')
    subprocess.Popen(['mm3d', 'GCPConvert', 'AppInFile',
                      os.path.join(out_dir, 'AutoGCPs{}.txt'.format(subscript))]).wait()

    print('writing AutoMeasures.txt')
    mmtools.write_auto_mesures(gcps, subscript, out_dir)

    print('running get_autogcp_locations.sh to get rough image locations for each point')
    subprocess.Popen(['get_autogcp_locations.sh', 'Ori-{}'.format(ori),
                      os.path.join(out_dir, 'AutoMeasures{}.txt'.format(subscript))] + imlist).wait()

    # print('searching for points in orthorectified images')
    print('finding image measures')
    mmtools.write_image_mesures(imlist, gcps, out_dir, subscript, ort_dir=ort_dir)

    print('running mm3d GCPBascule to estimate terrain errors')
    gcps = mmtools.run_bascule(gcps, out_dir, match_pattern, subscript, ori)
    gcps['res_dist'] = np.sqrt(gcps.xres**2 + gcps.yres**2)

    gcps = gcps[gcps.res_dist < 4 * nmad(gcps.res_dist)]

    mmtools.save_gcps(gcps, out_dir, utm_str, subscript)
    gcps = mmtools.run_bascule(gcps, out_dir, match_pattern, subscript, ori)
    gcps['res_dist'] = np.sqrt(gcps.xres**2 + gcps.yres**2)

    gcps = gcps[gcps.res_dist < 4 * nmad(gcps.res_dist)]

    mmtools.save_gcps(gcps, out_dir, utm_str, subscript)

    gcps = mmtools.run_bascule(gcps, out_dir, match_pattern, subscript, ori)
    gcps = mmtools.run_campari(gcps, out_dir, match_pattern, subscript, ref_img.dx, ortho_res)
    gcps['camp_dist'] = np.sqrt(gcps.camp_xres**2 + gcps.camp_yres**2)

    # if args.imgsource != 'DECLASSII':
    #     out = sliding_window_filter([ortho.shape[1], ortho.shape[0]], gcps, 2000, mindist=1000, how='camp_dist')
    #    gcps = gcps.loc[out]
    niter = 0
    while any([np.any(gcps.camp_res > 4 * nmad(gcps.camp_res)),
               np.any(gcps.camp_dist > 4 * nmad(gcps.camp_dist)),
               gcps.camp_res.max() > 2]) and niter <= 5:
        valid_inds = np.logical_and.reduce((gcps.camp_res < 4 * nmad(gcps.camp_res),
                                           gcps.camp_res < gcps.camp_res.max(),
                                           gcps.z_corr > gcps.z_corr.min()))
        if np.count_nonzero(valid_inds) < 10:
            break

        gcps = gcps.loc[valid_inds]
        mmtools.save_gcps(gcps, out_dir, utm_str, subscript)
        gcps = mmtools.run_bascule(gcps, out_dir, match_pattern, subscript, ori)
        gcps['res_dist'] = np.sqrt(gcps.xres ** 2 + gcps.yres ** 2)

        gcps = mmtools.run_campari(gcps, out_dir, match_pattern, subscript, ref_img.dx, ortho_res)
        gcps['camp_dist'] = np.sqrt(gcps.camp_xres**2 + gcps.camp_yres**2)
        niter += 1

    # final write of gcps to disk.
    gcps.to_file(os.path.join(out_dir, 'AutoGCPs{}.shp'.format(subscript)))

    fig1 = plt.figure(figsize=(7, 5))
    plt.imshow(ortho[::5, ::5], cmap='gray', extent=[0, ortho.shape[1], ortho.shape[0], 0])
    plt.plot(gcps.orig_j, gcps.orig_i, 'r+')
    plt.quiver(gcps.orig_j, gcps.orig_i, gcps.camp_xres, gcps.camp_yres, color='r')

    plt.savefig(os.path.join(out_dir, 'relative_gcps{}.png'.format(subscript)), bbox_inches='tight', dpi=200)
    plt.close(fig1)

    fig2 = ref_img.display(sfact=10, fig=plt.figure(figsize=(7, 5)))
    plt.plot(gcps.geometry.x, gcps.geometry.y, 'r+')
    plt.savefig(os.path.join(out_dir, 'world_gcps{}.png'.format(subscript)), bbox_inches='tight', dpi=200)
    plt.close(fig2)

    print('cleaning up.')
    # remove Auto-im.tif.txt, NoDist-im.tif.txt, etc. Ori-Relative-NoDist
    shutil.rmtree('Ori-{}-NoDist/'.format(ori))

    for txtfile in glob('Auto-OIS*.tif.txt') + \
                   glob('NoDist-OIS*.tif.txt'):
        os.remove(txtfile)
    print('end.')
    # embed()


def _search_size(imshape):
    min_dst = 100
    dst = min(400, imshape[0] / 10, imshape[1] / 10)
    return int(max(min_dst, dst))


def register_ortho(fn_ortho, fn_ref, fn_reldem, fn_dem, glacmask=None, landmask=None, footprints=None,
                   im_subset=None, block_num=None, ori='Relative', ortho_res=8.,
                   imgsource='DECLASSII', density=200, out_dir=None, allfree=True):
    """
    Register a relative orthoimage and DEM to a reference orthorectified image and DEM.

    :param str fn_ortho: path to relative orthoimage
    :param str fn_ref: path to reference orthorectified image
    :param str fn_reldem: path to relative DEM
    :param str fn_dem: path to reference DEM
    :param str glacmask: path to file of glacier outlines (i.e., an exclusion mask)
    :param str landmask: path to file of land outlines (i.e., an inclusion mask)
    :param str footprints: path to shapefile of image outlines. If not set, will download from USGS.
    :param str im_subset: subset of raw images to work with
    :param str block_num: block number to use if processing multiple image blocks
    :param str ori: name of orientation directory (after Ori-) (default: Relative)
    :param float ortho_res: approx. ground sampling distance (pixel resolution) of ortho image (default: 8 m)
    :param str imgsource: USGS dataset name for images (default: DECLASSII)
    :param int density: pixel spacing to look for GCPs (default: 200)
    :param str out_dir: output directory to save auto GCP files to (default: auto_gcps)
    :param bool allfree: run Campari setting all parameters free (default: True)
    """
    print('start.')

    if out_dir is None:
        out_dir = 'auto_gcps'

    mkdir_p(out_dir)

    if block_num is not None:
        subscript = '_block{}'.format(block_num)
    else:
        subscript = ''

    ort_dir = os.path.dirname(fn_ortho)

    imlist, match_pattern = get_imlist(im_subset)

    ref_img = GeoImg(fn_ref)
    # ref_lowres = ref_img.resample(init_res)
    # resamp_fact = init_res / ref_img.dx

    utm_str = get_utm_str(ref_img.epsg)

    ortho = imread(fn_ortho)
    # ortho_ = Image.fromarray(ortho)

    fn_tfw = fn_ortho.replace('.tif', '.tfw')
    with open(fn_tfw, 'r') as f:
        ortho_gt = [float(l.strip()) for l in f.readlines()]

    if footprints is None:
        clean_imlist = [im.split('OIS-Reech_')[-1].split('.tif')[0] for im in imlist]
        print('Attempting to get image footprints from USGS EarthExplorer.')
        footprints = get_usgs_footprints(clean_imlist, dataset=imgsource)
    else:
        footprints = gpd.read_file(footprints)

    mask_full, _, ref_img = get_mask(footprints, ref_img, imlist, landmask, glacmask)

    Minit, _, centers = transform_centers(ortho_gt, ref_img, imlist, footprints, 'Ori-{}'.format(ori))
    rough_tfm = warp(ortho, Minit, output_shape=ref_img.shape, preserve_range=True)

    fig, ax = plt.subplots(1, 2, figsize=(7, 5))
    ax[0].imshow(rough_tfm[::10, ::10], extent=[0, rough_tfm.shape[1], rough_tfm.shape[0], 0], cmap='gray')
    ax[1].imshow(ref_img.img[::10, ::10], extent=[0, ref_img.shape[1], ref_img.shape[0], 0], cmap='gray')

    plt.savefig('initial_transformation{}.png'.format(subscript), dpi=200, bbox_inches='tight')
    plt.close(fig)

    # for each of these pairs (src, dst), find the precise subpixel match (or not...)
    gcps = imtools.find_grid_matches(rough_tfm, ref_img, mask_full, Minit,
                                     spacing=density, dstwin=_search_size(rough_tfm.shape))

    xy = np.array([ref_img.ij2xy((pt[1], pt[0])) for pt in gcps[['search_j', 'search_i']].values]).reshape(-1, 2)
    gcps['geometry'] = [Point(pt) for pt in xy]

    gcps = gcps.loc[mask_full[gcps.search_i, gcps.search_j] == 255]

    gcps['rel_x'] = ortho_gt[4] + gcps['orig_j'].values * ortho_gt[0]  # need the original image coordinates
    gcps['rel_y'] = ortho_gt[5] + gcps['orig_i'].values * ortho_gt[3]

    gcps['elevation'] = 0
    gcps.set_crs(crs=ref_img.proj4, inplace=True)

    gcps.dropna(inplace=True)
    print('{} potential matches found'.format(gcps.shape[0]))

    print('loading dems')
    dem = GeoImg(fn_dem, dtype=np.float32)
    rel_dem = GeoImg(fn_reldem)

    # gcps.crs = {'init': 'epsg:{}'.format(ref_img.epsg)}
    for i, row in gcps.to_crs(crs=dem.proj4).iterrows():
        gcps.loc[i, 'elevation'] = dem.raster_points([(row.geometry.x, row.geometry.y)], nsize=3, mode='linear')
        if os.path.exists(fn_ortho.replace('.tif', '.tfw')):
            gcps.loc[i, 'el_rel'] = rel_dem.raster_points([(row.rel_x, row.rel_y)], nsize=3, mode='linear')

    # drop any gcps where we don't have a DEM value or a valid match
    gcps.loc[np.abs(gcps.elevation - dem.NDV) < 1, 'elevation'] = np.nan
    gcps.dropna(inplace=True)
    print('{} matches with valid elevations'.format(gcps.shape[0]))

    # run ransac to find the matches between the transformed image and the master image make a coherent transformation
    # residual_threshold is 10 pixels to allow for some local distortions, but get rid of the big blunders
    Mref, inliers_ref = ransac((gcps[['search_j', 'search_i']].values, gcps[['match_j', 'match_i']].values),
                               AffineTransform, min_samples=6, residual_threshold=10, max_trials=5000)

    gcps['aff_resid'] = Mref.residuals(gcps[['search_j', 'search_i']].values,
                                       gcps[['match_j', 'match_i']].values)

    out = sliding_window_filter([ortho.shape[1], ortho.shape[0]], gcps,
                                min(1000, ortho.shape[1] / 4, ortho.shape[0] / 4),
                                mindist=500, how='aff_resid', is_ascending=False)
    gcps = gcps.loc[out]

    print('{} valid matches found after estimating transformation'.format(gcps.shape[0]))

    gcps.index = range(gcps.shape[0])  # make sure index corresponds to row we're writing out
    gcps['id'] = ['GCP{}'.format(i) for i in range(gcps.shape[0])]
    gcps.to_file(os.path.join(out_dir, 'AutoGCPs{}.shp'.format(subscript)))

    print('writing AutoGCPs.txt')
    mmtools.write_auto_gcps(gcps, subscript, out_dir, utm_str)

    print('converting AutoGCPs.txt to AutoGCPs.xml')
    subprocess.Popen(['mm3d', 'GCPConvert', 'AppInFile',
                      os.path.join(out_dir, 'AutoGCPs{}.txt'.format(subscript))]).wait()

    print('writing AutoMeasures.txt')
    mmtools.write_auto_mesures(gcps, subscript, out_dir)

    print('running get_autogcp_locations.sh to get rough image locations for each point')
    subprocess.Popen(['get_autogcp_locations.sh', 'Ori-{}'.format(ori),
                      os.path.join(out_dir, 'AutoMeasures{}.txt'.format(subscript))] + imlist).wait()

    # print('searching for points in orthorectified images')
    print('finding image measures')
    mmtools.write_image_mesures(imlist, gcps, out_dir, subscript, ort_dir=ort_dir)

    print('running mm3d GCPBascule to estimate terrain errors')
    gcps = mmtools.run_bascule(gcps, out_dir, match_pattern, subscript, ori)
    gcps['res_dist'] = np.sqrt(gcps.xres ** 2 + gcps.yres ** 2)

    gcps = gcps[gcps.res_dist < 4 * nmad(gcps.res_dist)]

    mmtools.save_gcps(gcps, out_dir, utm_str, subscript)
    gcps = mmtools.run_bascule(gcps, out_dir, match_pattern, subscript, ori)
    gcps['res_dist'] = np.sqrt(gcps.xres ** 2 + gcps.yres ** 2)

    gcps = gcps[gcps.res_dist < 4 * nmad(gcps.res_dist)]

    mmtools.save_gcps(gcps, out_dir, utm_str, subscript)

    gcps = mmtools.run_bascule(gcps, out_dir, match_pattern, subscript, ori)
    gcps = mmtools.run_campari(gcps, out_dir, match_pattern, subscript, ref_img.dx, ortho_res, allfree=allfree)
    gcps['camp_dist'] = np.sqrt(gcps.camp_xres ** 2 + gcps.camp_yres ** 2)

    niter = 0
    while any([np.any(gcps.camp_res > 4 * nmad(gcps.camp_res)),
               np.any(gcps.camp_dist > 4 * nmad(gcps.camp_dist)),
               gcps.camp_res.max() > 2]) and niter <= 5:
        valid_inds = np.logical_and.reduce((gcps.camp_res < 4 * nmad(gcps.camp_res),
                                            gcps.camp_res < gcps.camp_res.max(),
                                            gcps.z_corr > gcps.z_corr.min()))
        if np.count_nonzero(valid_inds) < 10:
            break

        gcps = gcps.loc[valid_inds]
        mmtools.save_gcps(gcps, out_dir, utm_str, subscript)
        gcps = mmtools.run_bascule(gcps, out_dir, match_pattern, subscript, ori)
        gcps['res_dist'] = np.sqrt(gcps.xres ** 2 + gcps.yres ** 2)

        gcps = mmtools.run_campari(gcps, out_dir, match_pattern, subscript, ref_img.dx, ortho_res, allfree=allfree)
        gcps['camp_dist'] = np.sqrt(gcps.camp_xres ** 2 + gcps.camp_yres ** 2)
        niter += 1

    # final write of gcps to disk.
    gcps.to_file(os.path.join(out_dir, 'AutoGCPs{}.shp'.format(subscript)))

    fig1 = plt.figure(figsize=(7, 5))
    plt.imshow(ortho[::5, ::5], cmap='gray', extent=[0, ortho.shape[1], ortho.shape[0], 0])
    plt.plot(gcps.orig_j, gcps.orig_i, 'r+')
    plt.quiver(gcps.orig_j, gcps.orig_i, gcps.camp_xres, gcps.camp_yres, color='r')

    plt.savefig(os.path.join(out_dir, 'relative_gcps{}.png'.format(subscript)), bbox_inches='tight', dpi=200)
    plt.close(fig1)

    fig2 = ref_img.display(sfact=10, fig=plt.figure(figsize=(7, 5)))
    plt.plot(gcps.geometry.x, gcps.geometry.y, 'r+')
    plt.savefig(os.path.join(out_dir, 'world_gcps{}.png'.format(subscript)), bbox_inches='tight', dpi=200)
    plt.close(fig2)

    print('cleaning up.')
    # remove Auto-im.tif.txt, NoDist-im.tif.txt, etc. Ori-Relative-NoDist
    shutil.rmtree('Ori-{}-NoDist/'.format(ori))

    for txtfile in glob('Auto-OIS*.tif.txt') + \
                   glob('NoDist-OIS*.tif.txt'):
        os.remove(txtfile)
    print('end.')
    # embed()



#!/usr/bin/env python
import argparse
import numpy as np
from spymicmac.image import find_reseau_grid


def _argparser():
    _parser = argparse.ArgumentParser(description="Find Reseau marks in a scanned KH-9 Hexagon image.",
                                      formatter_class=argparse.RawDescriptionHelpFormatter)
    _parser.add_argument('fn_img', action='store', type=str, help='Image to find Reseau grid in.')
    _parser.add_argument('-csize', action='store', type=int, default=361, help='Reseau mark template size [361 pixels]')
    _parser.add_argument('-tsize', action='store', type=int, default=300, help='half-size of search window [300 pixels]')
    _parser.add_argument('-nproc', action='store', type=int, default=1,
                         help='Number of processors to use [1].')
    return _parser


def main():
    np.seterr(divide='ignore', invalid='ignore')
    parser = _argparser()
    args = parser.parse_args()

    find_reseau_grid(args.fn_img,
                     csize=args.csize,
                     tsize=args.tsize)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import argparse
from spymicmac.micmac import move_bad_tapas


def _argparser():
    parser = argparse.ArgumentParser(description="Read Tapas residuals and move images with NaN residual.",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('ori', action='store', type=str, help='Orientation directory to read.')

    return parser


def main():
    parser = _argparser()
    args = parser.parse_args()

    move_bad_tapas(args.ori)


if __name__ == "__main__":
    main()



"""
sPyMicMac.usgs is a collection of tools for interfacing with USGS Earth Explorer.
"""
import os
import netrc
import geopandas as gpd
from shapely.geometry.polygon import Polygon
from usgs import api


def get_login_creds():
    """
    Read a user's .netrc file and return the login credentials.

    :return:
        - **creds** -- the netrc.netrc credentials.
    """
    return netrc.netrc(os.path.expanduser('~/.netrc'))


def read_coords(result):
    """
    Parse a search result returned from USGS and create a list of coordinates for the image footprint.

    :param dict result: the USGS search result
    :return:
        -- **coords** (*list*) -- a list of coordinates
    """
    corner_names = ['NW', 'NE', 'SE', 'SW']
    corner_fields = [d for d in result['metadataFields'] if 'Corner' in d['fieldName'] and 'dec' in d['fieldName']]
    corner_dict = dict()

    for field in corner_fields:
        corner_dict[field['fieldName']] = float(field['value'])
    coords = []

    for corner in corner_names:
        coords.append((corner_dict['{} Corner Long dec'.format(corner)],
                       corner_dict['{} Corner Lat dec'.format(corner)]))

    return coords


def get_usgs_footprints(imlist, dataset='DECLASSII'):
    """
    Search for a list of images on USGS Earth Explorer. Note that the image names should be the USGS entity ID (e.g.,
    AR5840034159994 rather than AR5840034159994.tif).

    :param list imlist: a list of image names.
    :param str dataset: the USGS dataset name to search (default: DECLASSII).

    :return:
        -- **gdf** (*GeoDataFrame*) -- a GeoDataFrame of image footprints.
    """
    # air photos: 'AERIAL_COMBIN'
    creds = get_login_creds()

    gdf = gpd.GeoDataFrame()

    user, _, pwd = creds.authenticators('earthexplorer.usgs.gov')
    login = api.login(user, pwd)
    del user, pwd, creds

    if login['errorCode'] is not None:
        print('Error logging in to USGS EROS.')
        raise login['error']
    else:
        search_results = api.metadata(dataset,
                                      node='EE',
                                      entityids=imlist)
        for i, result in enumerate(search_results['data']):
            # coords = result['spatialFootprint']['coordinates'][0]
            coords = read_coords(result)
            poly = Polygon(coords)

            gdf.loc[i, 'geometry'] = poly
            gdf.loc[i, 'ID'] = result['entityId']

        gdf.crs = {'init': 'epsg:4326'}

        return gdf



"""
spymicmac.image is a collection of tools for working with aerial images.
"""
import os
import shutil
from glob import glob
import multiprocessing as mp
from functools import partial
from itertools import chain
import PIL.Image
import cv2
from osgeo import gdal
import matplotlib.pyplot as plt
import pandas as pd
import lxml.etree as etree
import lxml.builder as builder
from rtree import index
from skimage import exposure, transform, morphology, io, filters
from skimage.morphology import binary_closing, binary_dilation, disk
from skimage.measure import ransac
from skimage.feature import peak_local_max
from skimage.transform import warp, AffineTransform, EuclideanTransform
from scipy.interpolate import RectBivariateSpline as RBS
from scipy import ndimage
import numpy as np
from shapely.ops import cascaded_union
from shapely.geometry import LineString
import geopandas as gpd
# from llc import jit_filter_function
from numba import jit
from pybob.image_tools import match_hist, reshape_geoimg, create_mask_from_shapefile, nanmedian_filter
from pybob.bob_tools import mkdir_p
from pymmaster.mmaster_tools import orient_footprint
from spymicmac.micmac import get_im_meas, parse_im_meas


######################################################################################################################
# image filtering tools
######################################################################################################################
@jit
def nanstd(a):
    return np.nanstd(a)


def cross_template(shape, width=3):
    if isinstance(shape, int):
        rows = shape
        cols = shape
    else:
        rows, cols = shape
    half_r = int((rows - 1) / 2)
    half_c = int((cols - 1) / 2)
    half_w = int((width - 1) / 2)

    cross = np.zeros((rows, cols))
    cross[half_r - half_w - 1:half_r + half_w + 2:width + 1, :] = 2
    cross[:, half_c - half_w - 1:half_c + half_w + 2:width + 1] = 2

    cross[half_r - half_w:half_r + half_w + 1, :] = 1
    cross[:, half_c - half_w:half_c + half_w + 1] = 1
    return cross


def make_template(img, pt, half_size):
    nrows, ncols = img.shape
    row, col = np.round(pt).astype(int)
    left_col = max(col - half_size, 0)
    right_col = min(col + half_size, ncols)
    top_row = max(row - half_size, 0)
    bot_row = min(row + half_size, nrows)
    row_inds = [row - top_row, bot_row - row]
    col_inds = [col - left_col, right_col - col]
    template = img[top_row:bot_row + 1, left_col:right_col + 1].copy()
    return template, row_inds, col_inds


def find_match(img, template):
    img_eq = filters.rank.equalize(img, selem=morphology.disk(100))
    # res = cross_filter(img_eq, template)
    res = cv2.matchTemplate(img_eq, template, cv2.TM_CCORR_NORMED)
    i_off = (img.shape[0] - res.shape[0]) / 2
    j_off = (img.shape[1] - res.shape[1]) / 2
    minval, _, minloc, _ = cv2.minMaxLoc(res)
    # maxj, maxi = maxloc
    minj, mini = minloc
    sp_delx, sp_dely = get_subpixel(res)
    # sp_delx, sp_dely = 0, 0
    return res, mini + i_off + sp_dely, minj + j_off + sp_delx


def get_subpixel(res, how='min'):
    assert how in ['min', 'max'], "have to choose min or max"

    mgx, mgy = np.meshgrid(np.arange(-1, 1.01, 0.1), np.arange(-1, 1.01, 0.1), indexing='xy')  # sub-pixel mesh

    if how == 'min':
        peakval, _, peakloc, _ = cv2.minMaxLoc(res)
        mml_ind = 2
    else:
        _, peakval, _, peakloc = cv2.minMaxLoc(res)
        mml_ind = 3

    rbs_halfsize = 3  # size of peak area used for spline for subpixel peak loc
    rbs_order = 4  # polynomial order for subpixel rbs interpolation of peak location

    if ((np.array([n - rbs_halfsize for n in peakloc]) >= np.array([0, 0])).all()
            & (np.array([(n + rbs_halfsize) for n in peakloc]) < np.array(list(res.shape))).all()):
        rbs_p = RBS(range(-rbs_halfsize, rbs_halfsize + 1), range(-rbs_halfsize, rbs_halfsize + 1),
                    res[(peakloc[1] - rbs_halfsize):(peakloc[1] + rbs_halfsize + 1),
                    (peakloc[0] - rbs_halfsize):(peakloc[0] + rbs_halfsize + 1)],
                    kx=rbs_order, ky=rbs_order)

        b = rbs_p.ev(mgx.flatten(), mgy.flatten())
        mml = cv2.minMaxLoc(b.reshape(21, 21))
        # mgx,mgy: meshgrid x,y of common area
        # sp_delx,sp_dely: subpixel delx,dely
        sp_delx = mgx[mml[mml_ind][0], mml[mml_ind][1]]
        sp_dely = mgy[mml[mml_ind][0], mml[mml_ind][1]]
    else:
        sp_delx = 0.0
        sp_dely = 0.0
    return sp_delx, sp_dely


def highpass_filter(img):
    """
    Subtract a low-pass from an image, to return a highpass filter.

    :param array-like img: the image to filter.
    :return:
        - **highpass** (*array-like*) -- the highpass-filtered image
    """
    v = img.copy()
    v[np.isnan(img)] = 0
    vv = ndimage.gaussian_filter(v, 3)

    w = 0 * img.copy() + 1
    w[np.isnan(img)] = 0
    ww = ndimage.gaussian_filter(w, 3)

    tmplow = vv / ww
    tmphi = img - tmplow
    return tmphi


def splitter(img, nblocks, overlap=0):
    new_width = int(np.floor(img.shape[1]/nblocks[1]))
    new_height = int(np.floor(img.shape[0]/nblocks[0]))

    simages = []
    top_inds = []
    left_inds = []
    for i in range(nblocks[0]):
        this_col = []
        this_top = []
        this_left = []

        for j in range(nblocks[1]):
            lind = max(0, j*new_width-overlap)
            rind = min(img.shape[1], (j+1)*new_width + overlap)
            tind = max(0, i*new_height - overlap)
            bind = min(img.shape[0], (i+1)*new_height + overlap)
            this_col.append(img[tind:bind, lind:rind].copy())
            this_top.append(tind)
            this_left.append(lind)
        simages.append(this_col)
        top_inds.append(this_top)
        left_inds.append(this_left)

    return list(chain.from_iterable(simages)), list(chain.from_iterable(top_inds)), list(chain.from_iterable(left_inds))


def get_subimg_offsets(split, shape, overlap=0):
    ims_x = np.array([s.shape[1] - 2 * overlap for s in split])
    ims_y = np.array([s.shape[0] - 2 * overlap for s in split])

    rel_x = np.cumsum(ims_x.reshape(shape), axis=1)
    rel_y = np.cumsum(ims_y.reshape(shape), axis=0)

    rel_x = np.concatenate((np.zeros((shape[0], 1)), rel_x[:, :-1]), axis=1)
    rel_y = np.concatenate((np.zeros((1, shape[1])), rel_y[:-1, :]), axis=0)

    return rel_x.astype(int), rel_y.astype(int)


def stretch_image(img, scale=(0, 1), mult=255, outtype=np.uint8, mask=None):
    """
    Apply a linear stretch to an image by clipping and stretching to quantiles.

    :param array-like img: the image to stretch.
    :param tuple scale: a the minimum and maximum quantile to stretch to. (default: (0, 1) - the minimum/maximum values
        of the image)
    :param int|float mult: a multiplier to scale the result to. (default: 255)
    :param numpy.dtype outtype: the numpy datatype to return the stretched image as. (default: np.uint8)
    :param array-like mask: a mask of pixels to ignore when calculating quantiles.
    :return:
        - **stretched** (*array-like*) -- the stretched image.
    """
    if mask is None:
        maxval = np.nanquantile(img, max(scale))
        minval = np.nanquantile(img, min(scale))
    else:
        maxval = np.nanquantile(img[mask], max(scale))
        minval = np.nanquantile(img[mask], min(scale))

    img[img > maxval] = maxval
    img[img < minval] = minval

    return (mult * (img - minval) / (maxval - minval)).astype(outtype)


def contrast_enhance(fn_img, mask_value=None, qmin=0.02, qmax=0.98, gamma=1.25):
    """
    Enhance image contrast in a three-step process. First, the image is processed with a median filter to reduce
    noise. Next, a linear contrast stretch is applied, and finally, a gamma adjustment is applied.

    :param str fn_img: the image filename.
    :param int|float mask_value: a mask value to use when filtering the image.
    :param float qmin: the minimum quantile to use for the linear contrast stretch (default: 0.02)
    :param float qmax: the maximum quantile to use for the linear contrast stretch (default: 0.98)
    :param float gamma: the value to use for the gamma adjustment
    :return:
        - **enhanced** (*array-like*) -- the contrast-enhanced image.
    """
    img = io.imread(fn_img)
    if mask_value is not None:
        img = img.astype(np.float32)
        img[img == mask_value] = np.nan

    filt = nanmedian_filter(img, footprint=disk(3))

    stretch = stretch_image(filt, scale=(qmin, qmax))
    gamma = exposure.adjust_gamma(stretch, gamma=gamma)
    return gamma


def make_binary_mask(img, mult_value=255, erode=0, mask_value=0):
    """
    Create a binary mask for an image based on a given mask value. Values equal to mask_value will be given a value
    of 0 in the mask, while all other values will be set equal to mult_value.

    :param array-like img: the image to create a mask for.
    :param int|float mult_value: the value indicating a non-masked value (default: 255).
    :param int erode: the size of the erosion operation to apply (default: 0).
    :param int mask_value: the value to mask in the image (default: 0).
    :return:
        - **mask** (*array-like*) the binary mask.
    """
    _mask = mult_value * np.ones(img.shape, dtype=np.uint8)
    if np.isfinite(mask_value):
        _mask[img == mask_value] = 0
    else:
        _mask[np.isnan(img)] = 0

    if erode > 0:
        erode_mask = morphology.binary_erosion(_mask, selem=morphology.disk(erode))
        _mask[~erode_mask] = 0

    return _mask


def balance_image(img):
    """
    Apply contrast-limited adaptive histogram equalization (CLAHE) on an image, then apply a de-noising filter.

    :param array-like img: the image to balance.
    :return:
        - **img_filt** (*array-like*) -- the balanced, filtered image.
    """
    img_eq = (255 * exposure.equalize_adapthist(img)).astype(np.uint8)
    img_filt = filters.median(img_eq, selem=disk(1))
    return img_filt


######################################################################################################################
# GCP matching tools
######################################################################################################################
def keypoint_grid(img, spacing=25, size=10):
    _j = np.arange(0, img.shape[1], spacing)
    _i = np.arange(0, img.shape[0], spacing)
    _gridj, _gridi = np.meshgrid(_j, _i)
    _grid = np.concatenate((_gridj.reshape(-1, 1), _gridi.reshape(-1, 1)), axis=1)

    return [cv2.KeyPoint(pt[0], pt[1], size) for pt in _grid]


def get_dense_keypoints(img, mask, npix=100, nblocks=None, return_des=False):
    """
    Find ORB keypoints by dividing an image into smaller parts.

    :param array-like img: the image to use.
    :param array-like mask: a mask to use for keypoint generation.
    :param int npix: the block size (in pixels) to divide the image into.
    :param int nblocks: the number of blocks to divide the image into. If set, overrides value given by npix.
    :param bool return_des: return the keypoint descriptors, as well
    :return:
        - **keypoints** (*list*) -- a list of keypoint locations
        - **descriptors** (*list*) -- if requested, a list of keypoint descriptors.
    """
    orb = cv2.ORB_create()
    keypts = []
    if return_des:
        descriptors = []

    if nblocks is None:
        x_tiles = np.floor(img.shape[1] / npix).astype(int)
        y_tiles = np.floor(img.shape[0] / npix).astype(int)
    else:
        x_tiles = nblocks
        y_tiles = nblocks

    olap = int(max(0.25 * img.shape[1]/x_tiles, 0.25 * img.shape[0]/y_tiles))

    split_img, oy, ox = splitter(img, (y_tiles, x_tiles), overlap=olap)
    split_msk, _, _ = splitter(mask, (y_tiles, x_tiles), overlap=olap)

    # rel_x, rel_y = get_subimg_offsets(split_img, (y_tiles, x_tiles), overlap=olap)

    for i, img_ in enumerate(split_img):

        kp, des = orb.detectAndCompute(img_, mask=split_msk[i])
        if return_des:
            if des is not None:
                for ds in des:
                    descriptors.append(ds)

        for p in kp:
            p.pt = p.pt[0] + ox[i], p.pt[1] + oy[i]
            keypts.append(p)

    if return_des:
        return keypts, np.array(descriptors)
    else:
        return keypts


def get_footprint_overlap(fprints):
    """
    Return the area where image footprints overlap.

    :param GeoDataFrame fprints: a GeoDataFrame of image footprints
    :return:
        - **intersection** (*shapely.Polygon*) -- the overlapping area (cascaded union) of the images.
    """
    if fprints.shape[0] == 1:
        return fprints.geometry.values[0]

    idx = index.Index()

    for pos, row in fprints.iterrows():
        idx.insert(pos, row['geometry'].bounds)

    intersections = []
    for poly in fprints['geometry']:
        merged = cascaded_union([fprints.loc[pos, 'geometry'] for pos in idx.intersection(poly.bounds) if fprints.loc[pos, 'geometry'] != poly])
        intersections.append(poly.intersection(merged))

    intersection = cascaded_union(intersections)

    # return intersection.minimum_rotated_rectangle
    return intersection


def get_footprint_mask(shpfile, geoimg, filelist, fprint_out=False):
    """
    Return a footprint mask for an image.

    :param str|GeoDataFrame shpfile: a filename or a GeoDataFrame representation of the footprints.
    :param GeoImg geoimg: the image to create a mask for.
    :param list filelist: a list of image names to use to create the footprint mask.
    :param bool fprint_out: return the polygon representation of the footprint mask.
    :return:
        - **mask** (*array-like*) -- the footprint mask
        - **fprint** (*shapely.Polygon*) -- the footprint polygon, if requested.
    """
    imlist = [im.split('OIS-Reech_')[-1].split('.tif')[0] for im in filelist]
    if isinstance(shpfile, str):
        footprints_shp = gpd.read_file(shpfile)
        fp = footprints_shp[footprints_shp.ID.isin(imlist)].copy()
    else:
        fp = shpfile[shpfile.ID.isin(imlist)].copy()

    fprint = get_footprint_overlap(fp.to_crs(geoimg.proj4))

    tmp_gdf = gpd.GeoDataFrame(columns=['geometry'])
    tmp_gdf.loc[0, 'geometry'] = fprint
    tmp_gdf.crs = geoimg.proj4
    tmp_gdf.to_file('tmp_fprint.shp')

    maskout = create_mask_from_shapefile(geoimg, 'tmp_fprint.shp')

    for f in glob('tmp_fprint.*'):
        os.remove(f)
    if fprint_out:
        return maskout, fprint
    else:
        return maskout


def get_rough_geotransform(img1, img2, landmask=None, footmask=None,
                           stretch=(0.05, 0.95), equalize=False):
    """
    Search for a rough transformation between two images using ORB keypoint matching.

    :param array-like img1: the image to be transformed.
    :param array-like img2: the image to match to.
    :param array-like landmask: an inclusion mask for img2.
    :param array-like footmask: an exclusion mask for img2.
    :param tuple stretch: the minimum and maximum quantiles to stretch img2 to.
    :param bool equalize: apply CLAHE to img2.
    :return:
        - **img1_tfm** (*array-like*) -- the transformed input image.
        - **Minit** (*AffineTransform*) -- the estimated affine transformation between img1 and img2.
        - **dst_pts**, **src_pts**, **inliers** (*array-like*) -- the matched points for img2, img1, and the inliers used to estimate Minit.
    """
    if equalize:
        img2_eq = (255 * exposure.equalize_adapthist(img2.img.astype(np.uint16),
                                                     clip_limit=0.03)).astype(np.uint8)
    else:
        img2_eq = stretch_image(img2.img, scale=stretch)

    img1_mask = make_binary_mask(img1, erode=3, mask_value=0)

    img2_mask = make_binary_mask(img2.img, erode=3, mask_value=np.nan)

    if landmask is not None:
        # lm = create_mask_from_shapefile(img2, landmask)
        img2_mask[~landmask] = 0

    if footmask is not None:
        # mask_ = get_footprint_mask(footmask, img2, imlist)
        img2_mask[~footmask] = 0

    kp, des, matches = get_matches(img1.astype(np.uint8), img2_eq,
                                   mask1=img1_mask, mask2=img2_mask, dense=True)
    print('{} matches found.'.format(len(matches)))

    src_pts = np.array([kp[0][m.queryIdx].pt for m in matches])
    dst_pts = np.array([kp[1][m.trainIdx].pt for m in matches])

    Minit, inliers = ransac((dst_pts, src_pts), AffineTransform,
                            min_samples=5, residual_threshold=10, max_trials=1000)
    print('{} matches used for initial transformation'.format(np.count_nonzero(inliers)))

    img1_tfm = warp(img1, Minit, output_shape=img2_eq.shape, preserve_range=True)

    return img1_tfm, Minit, (dst_pts, src_pts, inliers)


def get_initial_transformation(img1, img2, pRes=800, landmask=None, footmask=None, imlist=None):
    im2_lowres = reshape_geoimg(img2, pRes, pRes)

    im2_eq = match_hist(im2_lowres.img, np.array(img1))
    im1_mask = 255 * np.ones(img1.shape, dtype=np.uint8)
    im1_mask[img1 == 0] = 0  # nodata from ortho

    im2_mask = 255 * np.ones(im2_eq.shape, dtype=np.uint8)
    im2_mask[im2_eq == 0] = 0

    if imlist is None:
        imlist = glob('OIS*.tif')

    if landmask is not None:
        mask_ = create_mask_from_shapefile(im2_lowres, landmask)
        im2_mask[~mask_] = 0
    if footmask is not None:
        mask_ = get_footprint_mask(footmask, im2_lowres, imlist)
        im2_mask[~mask_] = 0

    kp, des, matches = get_matches(img1, im2_eq, mask1=im1_mask, mask2=im2_mask)
    src_pts = np.array([kp[0][m.queryIdx].pt for m in matches])
    dst_pts = np.array([kp[1][m.trainIdx].pt for m in matches])
    # aff_matrix, good_mask = cv2.estimateAffine2D(src_pts, dst_pts, ransacReprojThreshold=25)
    Mout, inliers = ransac((dst_pts, src_pts), EuclideanTransform,
                           min_samples=5, residual_threshold=2, max_trials=1000)
    # check that the transformation was successful by correlating the two images.
    # im1_tfm = cv2.warpAffine(img1, aff_matrix, (im2_lowres.img.shape[1], im2_lowres.img.shape[0]))
    im1_tfm = warp(img1, Mout, output_shape=im2_lowres.img.shape, preserve_range=True)
    im1_pad = np.zeros(np.array(im1_tfm.shape) + 2, dtype=np.uint8)
    im1_pad[1:-1, 1:-1] = im1_tfm
    res = cv2.matchTemplate(np.ma.masked_values(im2_eq, 0),
                            np.ma.masked_values(im1_pad, 0),
                            cv2.TM_CCORR_NORMED)
    print(res[1, 1])
    success = res[1, 1] > 0.5

    return Mout, success, im2_eq.shape


def get_matches(img1, img2, mask1=None, mask2=None, dense=False):

    if dense:
        if np.any(np.array(img1.shape) < 100) or np.any(np.array(img2.shape) < 100):
            kp1, des1 = get_dense_keypoints(img1.astype(np.uint8), mask1, nblocks=1, return_des=True)
            kp2, des2 = get_dense_keypoints(img2.astype(np.uint8), mask2, nblocks=1, return_des=True)
        else:
            kp1, des1 = get_dense_keypoints(img1.astype(np.uint8), mask1, return_des=True)
            kp2, des2 = get_dense_keypoints(img2.astype(np.uint8), mask2, return_des=True)
    else:
        orb = cv2.ORB_create()
        kp1, des1 = orb.detectAndCompute(img1.astype(np.uint8), mask=mask1)
        kp2, des2 = orb.detectAndCompute(img2.astype(np.uint8), mask=mask2)

    flann_idx = 6
    index_params = dict(algorithm=flann_idx, table_number=12, key_size=20, multi_probe_level=2)
    search_params = dict(checks=10000)

    flann = cv2.FlannBasedMatcher(index_params, search_params)
    raw_matches = flann.knnMatch(des1, des2, k=2)
    matches = []
    for m in raw_matches:
        if len(m) == 2 and m[0].distance < m[1].distance * 0.75:
            matches.append(m[0])

    return (kp1, kp2), (des1, des2), matches


def find_gcp_match(img, template, method=cv2.TM_CCORR_NORMED):
    res = cv2.matchTemplate(img, template, method)
    i_off = (img.shape[0] - res.shape[0]) / 2
    j_off = (img.shape[1] - res.shape[1]) / 2
    _, maxval, _, maxloc = cv2.minMaxLoc(res)
    maxj, maxi = maxloc
    try:
        sp_delx, sp_dely = get_subpixel(res, how='max')
    except ValueError as e:
        sp_delx = 0
        sp_dely = 0

    return res, maxi + i_off + sp_dely, maxj + j_off + sp_delx


def find_grid_matches(tfm_img, refgeo, mask, initM=None, spacing=200, srcwin=60, dstwin=600):
    """
    Find matches between two images on a grid using normalized cross-correlation template matching.

    :param array-like tfm_img: the image to use for matching.
    :param GeoImg refgeo: the reference image to use for matching.
    :param array-like mask: a mask indicating areas that should be used for matching.
    :param initM: the model used for transforming the initial, non-georeferenced image.
    :param int spacing: the grid spacing, in pixels (default: 200 pixels)
    :param int srcwin: the half-size of the template window.
    :param int dstwin: the half-size of the search window.
    :return:
        - **gcps** (*pandas.DataFrame*) -- a DataFrame with GCP locations, match strength, and other information.
    """
    match_pts = []
    z_corrs = []
    peak_corrs = []
    res_imgs = []

    jj = np.arange(srcwin, spacing * np.ceil((refgeo.img.shape[1]-srcwin) / spacing) + 1, spacing).astype(int)
    ii = np.arange(srcwin, spacing * np.ceil((refgeo.img.shape[0]-srcwin) / spacing) + 1, spacing).astype(int)

    search_pts = []

    for _i in ii:
        for _j in jj:
            search_pts.append((_j, _i))
            # for pt in search_pts:
            # if mask[pt[1], pt[0]] == 0:
            submask, _, _ = make_template(mask, (_i, _j), srcwin)
            if np.count_nonzero(submask) / submask.size < 0.05:
                match_pts.append((-1, -1))
                z_corrs.append(np.nan)
                peak_corrs.append(np.nan)
                res_imgs.append(np.nan)
                continue
            try:
                testchip, _, _ = make_template(refgeo.img, (_i, _j), srcwin)
                dst_chip, _, _ = make_template(tfm_img, (_i, _j), dstwin)

                testchip[np.isnan(testchip)] = 0
                dst_chip[np.isnan(dst_chip)] = 0

                test = highpass_filter(testchip)
                dest = highpass_filter(dst_chip)

                testmask = binary_dilation(testchip == 0, selem=disk(8))
                destmask = binary_dilation(dst_chip == 0, selem=disk(8))

                test[testmask] = np.random.rand(test.shape[0], test.shape[1])[testmask]
                dest[destmask] = np.random.rand(dest.shape[0], dest.shape[1])[destmask]

                corr_res, this_i, this_j = find_gcp_match(dest.astype(np.float32), test.astype(np.float32))
                peak_corr = cv2.minMaxLoc(corr_res)[1]

                pks = peak_local_max(corr_res, min_distance=5, num_peaks=2)
                this_z_corrs = []
                for pk in pks:
                    max_ = corr_res[pk[0], pk[1]]
                    this_z_corrs.append((max_ - corr_res.mean()) / corr_res.std())
                dz_corr = max(this_z_corrs) / min(this_z_corrs)
                z_corr = max(this_z_corrs)

                # if the correlation peak is very high, or very unique, add it as a match
                out_i, out_j = this_i - dstwin + _i, this_j - dstwin + _j
                z_corrs.append(z_corr)
                peak_corrs.append(peak_corr)
                match_pts.append([out_j, out_i])
                res_imgs.append(corr_res)
            except:
                match_pts.append((-1, -1))
                z_corrs.append(np.nan)
                peak_corrs.append(np.nan)
                res_imgs.append(np.nan)

    search_pts = np.array(search_pts)
    _dst = np.array(match_pts)

    gcps = gpd.GeoDataFrame()
    gcps['pk_corr'] = peak_corrs
    gcps['z_corr'] = z_corrs
    gcps['match_j'] = _dst[:, 0]  # points matched in transformed image
    gcps['match_i'] = _dst[:, 1]

    if initM is not None:
        # have to find the initial, which means back-transforming match_pts with Minit_full
        # _src = np.dot(initM.params, np.hstack([_dst, np.ones(_dst[:, 0].shape).reshape(-1, 1)]).T).T[:, :2]
        _src = initM(_dst)
        gcps['orig_j'] = _src[:, 0]  # this should be the back-transformed match_pts
        gcps['orig_i'] = _src[:, 1]

    gcps['search_j'] = search_pts[:, 0]
    gcps['search_i'] = search_pts[:, 1]
    gcps['dj'] = gcps['search_j'] - gcps['match_j']
    gcps['di'] = gcps['search_i'] - gcps['match_i']

    gcps.dropna(inplace=True)

    return gcps


def transform_from_fprint(img, geo, fprint):
    """
    Using a georeferenced footprint for an image, estimate a Euclidean transform.

    :param array-like img: the image to transform.
    :param GeoImg geo: the reference image to transform the input image to.
    :param shapely.Polygon fprint: the image footprint to use for transformation.
    :return:
        - **transform** (*GeometricTransform*) -- the estimated euclidean transform for the input image.
    """
    oprint = orient_footprint(fprint)
    h, w = img.shape
    src_pts = np.array([[0, 0], [0, h-1], [w-1, h-1], [w-1, 0]]).reshape(-1, 2)
    x, y = oprint.boundary.xy
    ij = [geo.xy2ij(pt) for pt in zip(x, y)]
    i = [pt[0] for pt in ij]
    j = [pt[1] for pt in ij]
    inds = np.array([1, 0, 3, 2]).reshape(-1, 1)
    dst_pts = np.array(list(zip(np.array(j)[inds], np.array(i)[inds]))).reshape(-1, 2)

    return transform.estimate_transform('euclidean', dst_pts, src_pts)


def find_cross(img, pt, cross, tsize=300):
    """
    Find the center of a Reseau mark (cross, or '+' shape).

    :param array-like img: the image to use.
    :param tuple pt: the initial search location.
    :param array-like cross: the Reseau mark template to use.
    :param int tsize: the search grid half-size (default: 300 -> 601x601)
    :return:
        - **center_i** (*float*) -- the row coordinate of the cross center.
        - **center_j** (*float*) -- the column coordinate of the cross center.
    """
    subimg, _, _ = make_template(img, pt, tsize)
    res, this_i, this_j = find_match(subimg, cross)
    inv_res = res.max() - res

    pks = peak_local_max(inv_res, min_distance=5, num_peaks=2)
    this_z_corrs = []
    for pk in pks:
        max_ = inv_res[pk[0], pk[1]]
        this_z_corrs.append((max_ - inv_res.mean()) / inv_res.std())
    if max(this_z_corrs) > 4 and max(this_z_corrs)/min(this_z_corrs) > 1.15:
        return this_i-tsize+pt[0], this_j-tsize+pt[1]
    else:
        return np.nan, np.nan


def lsq_fit(x, y, z):
    tmp_A = []
    tmp_b = []

    for kk in range(z.size):
        _ind = np.unravel_index(kk, z.shape)
        tmp_A.append([x[_ind], y[_ind], 1])
        tmp_b.append(z[_ind])

    A = np.matrix(tmp_A)
    b = np.matrix(tmp_b).T

    A = A[np.isfinite(z.reshape(-1, 1)).any(axis=1)]
    b = b[np.isfinite(b)]

    fit = (A.T * A).I * A.T * b.T
    out_lsq = np.zeros(z.shape)
    for kk in range(z.size):
        _ind = np.unravel_index(kk, z.shape)
        out_lsq[_ind] = fit[0] * x[_ind] + fit[1] * y[_ind] + fit[2]
    return out_lsq


def _moving_average(a, n=5):
    ret = np.cumsum(a)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n


def get_perp(corners):
    a = np.array([corners[1][0] - corners[0][0],
                  corners[1][1] - corners[0][1]])
    a = a / np.linalg.norm(a)
    return np.array([-a[1], a[0]])


def find_reseau_border(img, rough_ext, cross, tsize=300):
    """
    Find the rough edges of the Reseau field in a KH-9 image.

    :param array-like img: the image to use.
    :param array-like rough_ext: the rough location (left, right, top, bottom) of the image border.
    :param array-like cross: a Reseau mark template.
    :param int tsize: the search grid size for the Reseau mark.
    :return:
        - **left_edge** (*LineString*) -- the left-side Reseau field border
        - **right_edge** (*LineString*) -- the right-side Reseau field border
    """
    left, right, top, bot = rough_ext

    if np.isnan(right):
        search_corners = [(bot-625, left+180), (top+625, left+180)]
    else:
        search_corners = [(bot-625, right-180), (top+625, right-180)]

    grid_corners = []
    for c in search_corners:
        _i, _j = find_cross(img, c, cross, tsize)
        if any(np.isnan([_j, _i])):
            grid_corners.append((c[1], c[0]))
        else:
            grid_corners.append((_j, _i))
    pixres = (grid_corners[0][1] - grid_corners[1][1]) / 22
    npix = 23 * pixres

    perp = get_perp(grid_corners)

    if np.isnan(right):
        left_edge = LineString([grid_corners[0], grid_corners[1]])
        right_edge = LineString([grid_corners[0] + npix * perp,
                                 grid_corners[1] + npix * perp])
    else:
        right_edge = LineString([grid_corners[0], grid_corners[1]])
        left_edge = LineString([grid_corners[0] - npix * perp,
                                grid_corners[1] - npix * perp])

    return left_edge, right_edge


def downsample_image(img, fact=4):
    _img = PIL.Image.fromarray(img)
    return np.array(_img.resize((np.array(_img.size) / fact).astype(int), PIL.Image.LANCZOS))


def get_rough_frame(img):
    """
    Find the rough location of an image frame/border.

    :param array-like img: the image to find a border for
    :return:
        - **xmin**, **xmax**, **ymin**, **ymax** (*float*) -- the left, right, top, and bottom indices for the rough border.
    """
    img_lowres = downsample_image(img, fact=10)
    img_seg = np.zeros(img_lowres.shape)
    img_seg[img_lowres > filters.threshold_local(img_lowres, 101)] = 1
    img_seg = binary_closing(img_seg, selem=disk(1))

    v_sob = filters.sobel_v(img_seg)**2
    h_sob = filters.sobel_h(img_seg)**2

    vert = np.count_nonzero(v_sob > 0.5, axis=0)
    hori = np.count_nonzero(h_sob > 0.5, axis=1)

    vert_thresh = 0.3 * vert.max()
    hori_thresh = 0.3 * hori.max()

    try:
        xmin = 10 * peak_local_max(vert[:200], num_peaks=2, min_distance=20,
                                   threshold_abs=vert_thresh, exclude_border=10).max()
    except ValueError:
        xmin = np.nan

    try:
        xmax = 10 * (peak_local_max(vert[-200:], num_peaks=2, min_distance=20,
                                    threshold_abs=vert_thresh, exclude_border=10).min() + img_lowres.shape[1] - 200)
    except ValueError:
        xmax = np.nan

    try:
        ymin = 10 * peak_local_max(hori[:200], num_peaks=2, min_distance=10,
                                   threshold_abs=hori_thresh, exclude_border=5).max()
    except ValueError:
        ymin = np.nan

    try:
        ymax = 10 * (peak_local_max(hori[-200:], num_peaks=2, min_distance=10,
                                    threshold_abs=hori_thresh, exclude_border=5).min() + img_lowres.shape[0] - 200)
    except ValueError:
        ymax = np.nan

    return xmin, xmax, ymin, ymax


def _search_grid(left, right):
    i_grid = []
    j_grid = []
    search_pts = []

    for ii in range(0, 23):
        this_left = left.interpolate(ii * left.length / 22)
        this_right = right.interpolate(ii * right.length / 22)
        this_line = LineString([this_left, this_right])
        for jj in range(0, 24):
            this_pt = this_line.interpolate(jj * this_line.length / 23)
            i_grid.append(ii)
            j_grid.append(jj)
            search_pts.append((this_pt.y, this_pt.x))

    return i_grid, j_grid, search_pts


def find_reseau_grid(fn_img, csize=361, tsize=300, nproc=1, return_val=False):
    """
    Find the locations of the Reseau marks in a scanned KH-9 image. Locations are saved
    to Ori-InterneScan/MeasuresIm-:fn_img:.xml.

    :param str fn_img: the image filename.
    :param int csize: the size of the cross template (default: 361 -> 361x361)
    :param int tsize: the search grid size for the Reseau mark.
    :param int nproc: the number of processors to use, via multiprocessing.Pool (default: 1).
    :param bool return_val: return a pandas DataFrame of the Reseau mark locations (default: False).
    :return:
        - **gcps_df** (*pandas.DataFrame*) -- a DataFrame of the Reseau mark locations (if return_val=True).
    """
    print('Reading {}'.format(fn_img))
    img = io.imread(fn_img)
    img_ = PIL.Image.fromarray(img)
    img_lowres = np.array(img_.resize((np.array(img_.size)/10).astype(int), PIL.Image.LANCZOS))
    del img_

    print('Image read.')
    tmp_cross = cross_template(csize, width=3)
    cross = np.ones((csize, csize)).astype(np.uint8)
    cross[tmp_cross == 1] = 255

    fig = plt.figure(figsize=(7, 12))
    ax = fig.add_subplot(111)
    ax.imshow(img_lowres, cmap='gray', extent=[0, img.shape[1], img.shape[0], 0])
    ax.set_xticks([])
    ax.set_yticks([])

    rough_border = get_rough_frame(img)
    left_edge, right_edge = find_reseau_border(img, rough_border, cross, tsize=tsize)

    II, JJ, search_grid = _search_grid(left_edge, right_edge)
    matched_grid = []

    print('Finding grid points in {}...'.format(fn_img))
    if nproc > 1:
        subimgs = []
        std_devs = []
        means = []
        for loc in search_grid:
            subimg, _, _ = make_template(img, loc, tsize)
            subimgs.append(subimg)
        pool = mp.Pool(nproc)
        outputs = pool.map(partial(find_match, template=cross), subimgs)
        pool.close()
        # have to map outputs to warped_ij
        for n, output in enumerate(outputs):
            res, this_i, this_j = output
            maxj, maxi = cv2.minMaxLoc(res)[2]  # 'peak' is actually the minimum location, remember.
            inv_res = res.max() - res
            std_devs.append(inv_res.std())
            means.append(inv_res.mean())

            pks = peak_local_max(inv_res, min_distance=5, num_peaks=2)
            if pks.size > 0:
                this_z_corrs = []
                for pk in pks:
                    max_ = inv_res[pk[0], pk[1]]
                    this_z_corrs.append((max_ - inv_res.mean()) / inv_res.std())

                if max(this_z_corrs) > 5 and max(this_z_corrs) / min(this_z_corrs) > 1.15:
                    rel_i = this_i - tsize
                    rel_j = this_j - tsize
                    i, j = search_grid[n]
                    matched_grid.append((rel_i + i, rel_j + j))
                else:
                    matched_grid.append((np.nan, np.nan))
            else:
                matched_grid.append((np.nan, np.nan))
    else:
        for loc in search_grid:
            _j, _i = find_cross(img, loc, cross, tsize=tsize)
            matched_grid.append((_j, _i))

    matched_grid = np.array(matched_grid)
    search_grid = np.array(search_grid)

    gcps_df = pd.DataFrame()
    for i, pr in enumerate(list(zip(II, JJ))):
        gcps_df.loc[i, 'gcp'] = 'GCP_{}_{}'.format(pr[0], pr[1])

    gcps_df['search_j'] = search_grid[:, 1]
    gcps_df['search_i'] = search_grid[:, 0]
    gcps_df['match_j'] = matched_grid[:, 1]
    gcps_df['match_i'] = matched_grid[:, 0]

    model = AffineTransform()
    model.estimate(gcps_df[['search_j', 'search_i']].values, gcps_df[['match_j', 'match_i']].values)
    dst = model(gcps_df[['search_j', 'search_i']].values)

    nomatch = np.isnan(gcps_df.match_j)

    ux = lsq_fit(gcps_df.search_j.values, gcps_df.search_i.values, gcps_df.match_j.values)
    uy = lsq_fit(gcps_df.search_j.values, gcps_df.search_i.values, gcps_df.match_i.values)

    gcps_df.loc[nomatch, 'match_j'] = ux[nomatch]
    gcps_df.loc[nomatch, 'match_i'] = uy[nomatch]

    gcps_df['dj'] = gcps_df['match_j'] - dst[:, 0]
    gcps_df['di'] = gcps_df['match_i'] - dst[:, 1]

    gcps_df['im_row'] = gcps_df['match_i']
    gcps_df['im_col'] = gcps_df['match_j']

    print('Grid points found.')
    mkdir_p('match_imgs')

    ax.quiver(gcps_df.search_j, gcps_df.search_i, gcps_df.dj, gcps_df.di, color='r')
    ax.plot(gcps_df.search_j[nomatch], gcps_df.search_i[nomatch], 'b+')
    this_out = os.path.splitext(fn_img)[0]
    fig.savefig(os.path.join('match_imgs', this_out + '_matches.png'), bbox_inches='tight', dpi=200)

    E = builder.ElementMaker()
    ImMes = E.MesureAppuiFlottant1Im(E.NameIm(fn_img))

    pt_els = get_im_meas(gcps_df, E)
    for p in pt_els:
        ImMes.append(p)
    mkdir_p('Ori-InterneScan')

    outxml = E.SetOfMesureAppuisFlottants(ImMes)
    tree = etree.ElementTree(outxml)
    tree.write(os.path.join('Ori-InterneScan', 'MeasuresIm-' + fn_img + '.xml'), pretty_print=True,
               xml_declaration=True, encoding="utf-8")

    if return_val:
        return gcps_df


def _fix_cross(subimg):
    subimg = subimg.astype(np.float32)

    cross = cross_template(subimg.shape[0], width=5)
    cross[:, :16] = 0
    cross[:, -16:] = 0
    cross[:16, :] = 0
    cross[-16:, :] = 0
    if subimg.shape[0] != cross.shape[0]:
        cross = cross[:subimg.shape[0], :]

    if subimg.shape[1] != cross.shape[1]:
        cross = cross[:, :subimg.shape[1]]

    subimg[cross != 0] = np.nan
    fixed = nanmedian_filter(subimg, footprint=disk(7))
    subimg[np.isnan(subimg)] = fixed[np.isnan(subimg)]
    return subimg.astype(np.uint8)


def remove_crosses(fn_img):
    """
    Remove the Reseau marks from a KH-9 image before re-sampling.

    :param str fn_img: the image filename.
    """
    fn_meas = os.path.join('Ori-InterneScan', 'MeasuresIm-{}.xml'.format(fn_img))
    img = io.imread(fn_img)
    gcps = parse_im_meas(fn_meas)

    pt = np.round([gcps.i[0], gcps.j[0]]).astype(int)
    subim, row_, col_ = make_template(img, pt, 200)
    cross = cross_template(subim.shape[0], width=5)
    cross[:, 16:22] = 1
    cross[:, -24:-16] = 1
    cross[16:22, :] = 1
    cross[-24:-16, :] = 1

    subim = subim.astype(np.float32)
    subim[cross != 0] = np.nan
    fixed = nanmedian_filter(subim, footprint=disk(7))
    subim[np.isnan(subim)] = fixed[np.isnan(subim)]
    img[int(pt[0]) - row_[0]:int(pt[0]) + row_[1] + 1, int(pt[1]) - col_[0]:int(pt[1]) + col_[1] + 1] = subim.astype(
        np.uint8)

    for i, row in gcps.loc[1:].iterrows():
        pt = np.round([row.i, row.j]).astype(int)
        subim, row_, col_ = make_template(img, pt, 200)
        img[pt[0] - row_[0]:pt[0] + row_[1] + 1, pt[1] - col_[0]:pt[1] + col_[1] + 1] = _fix_cross(subim)

    mkdir_p('original')
    shutil.move(fn_img, 'original')
    io.imsave(fn_img, img.astype(np.uint8))


def join_hexagon(im_pattern, overlap=2000, blend=True, corona=False):
    """
    Join two halves of a scanned KH-9 Hexagon image (or four parts of a scanned KH-4 Corona image).

    :param str im_pattern: the base name of the image to use (e.g., DZB1216-500280L002001).
    :param int overlap: the overlap, in pixels, between the image parts.
    :param bool blend: apply a linear blend between the two scanned halves (default: True).
    :param bool corona: image is a KH-4/4A Corona image. If True, looks for four image parts instead
        of two (default: False).
    """
    if not corona:
        left = io.imread('{}_a.tif'.format(im_pattern))
        right = io.imread('{}_b.tif'.format(im_pattern))

        left_gd = gdal.Open('{}_a.tif'.format(im_pattern))
        right_gd = gdal.Open('{}_b.tif'.format(im_pattern))

        M = match_halves(left, right, overlap=overlap)

        out_shape = (left.shape[0], left.shape[1] + right.shape[1])

        combined_right = warp(right, M, output_shape=out_shape, preserve_range=True, order=3)

        combined_left = np.zeros(out_shape)
        combined_left[:, :left.shape[1]] = left

        if blend:
            combined = _blend(combined_left, combined_right, left.shape)
        else:
            combined = combined_left + combined_right

        last_ind = np.where(np.sum(combined, axis=0) > 0)[0][-1]
        combined = combined[:, :last_ind+1]

        io.imsave('{}.tif'.format(im_pattern), combined.astype(np.uint8))
    else:
        left = io.imread('{}_d.tif'.format(im_pattern))
        for right_img in ['c', 'b', 'a']:
            right = io.imread('{}_{}.tif'.format(im_pattern, right_img))

            M = match_halves(left, right, overlap=overlap)
            out_shape = (left.shape[0], left.shape[1] + right.shape[1])

            combined_right = warp(right, M, output_shape=out_shape, preserve_range=True, order=3)

            combined_left = np.zeros(out_shape)
            combined_left[:, :left.shape[1]] = left

            if blend:
                combined = _blend(combined_left, combined_right, left.shape)
            else:
                combined = combined_left + combined_right

            last_ind = np.where(np.sum(combined, axis=0) > 0)[0][-1]
            left = combined[:, :last_ind + 1]

        io.imsave('{}.tif'.format(im_pattern), combined.astype(np.uint8))


def _blend(_left, _right, left_shape):
    first = np.where(np.sum(_right, axis=0) > 0)[0][0]
    last = left_shape[1]

    m = 1 / (first - last)
    alpha = np.ones(_left.shape, dtype=np.float32)
    alpha[:, last:] = 0
    for i, ind in enumerate(np.arange(first, last)):
        alpha[:, ind] = 1 + m * (ind - first)

    return alpha * _left + (1 - alpha) * _right


def match_halves(left, right, overlap):
    """
    Find a transformation to join the left and right halves of an image scan.

    :param array-like left: the left-hand image scan.
    :param array-like right: the right-hand image scan.
    :param int overlap: the estimated overlap between the two images, in pixels.
    :return:
        - **model** (*EuclideanTransform*) -- the estimated Euclidean transformation between the two image halves.
    """
    src_pts = []
    dst_pts = []

    row_inds = list(range(0, left.shape[0] + 1, overlap))
    if row_inds[-1] != left.shape[0]:
        row_inds.append(-1)
    row_inds = np.array(row_inds)

    for i, ind in enumerate(row_inds[:-1]):
        try:
            l_sub = left[ind:row_inds[i + 1], -overlap:]
            r_sub = right[ind:row_inds[i + 1], :overlap]

            kp, des, matches = get_matches(l_sub, r_sub)
            src_pts.extend(
                [np.array(kp[0][m.queryIdx].pt) + np.array([left.shape[1] - overlap, ind]) for m in matches])
            dst_pts.extend([np.array(kp[1][m.trainIdx].pt) + np.array([0, ind]) for m in matches])
        except:
            continue

    model, inliers = ransac((np.array(src_pts), np.array(dst_pts)), EuclideanTransform,
                        min_samples=25, residual_threshold=1, max_trials=1000)
    print('{} tie points found'.format(np.count_nonzero(inliers)))
    return model



from pathlib import Path
from setuptools import find_packages, setup

readme = Path(__file__).parent / 'README.md'

setup(name='spymicmac',
      version='0.1.1',
      description='a python package for processing KH-9 and historical aerial imagery using MicMac',
      long_description=readme.read_text(),
      long_description_content_type='text/markdown',
      url='https://github.com/iamdonovan/sPyMicMac',
      doc_url='https://spymicmac.readthedocs.io/',
      author='Bob McNabb',
      author_email='robertmcnabb@gmail.com',
      maintainer='iamdonovan',
      license='GPL-3.0',
      license_file='LICENSE',
      include_package_data=True,
      classifiers=[
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Natural Language :: English',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Unix Shell',
          'Topic :: Scientific/Engineering :: GIS',
          'Topic :: Scientific/Engineering :: Image Processing'
      ],
      python_requires='>=3.7',
      install_requires=['numpy', 'scipy', 'matplotlib', 'lxml', 'pybob>0.25',
                        'shapely', 'opencv-python', 'pandas', 'geopandas', 'pymmaster>0.1',
                        'scikit-image>=0.18', 'gdal>=3.2.0',
                        'sphinx-argparse', 'earthengine-api', 'pyasn1', 'usgs'],
      packages=find_packages(),
      scripts=['bin/get_autogcp_locations.sh',
               'bin/post_process_micmac.sh',
               'bin/resample_hexagon.sh'],
      entry_points={
          'console_scripts': [
              'balance_images = spymicmac.tools.balance_images:main',
              'combine_auto_measures = spymicmac.tools.combine_auto_measures:main',
              'find_reseau_grid = spymicmac.tools.find_reseau_grid:main',
              'generate_micmac_measures = spymicmac.tools.generate_micmac_measures:main',
              'join_hexagon = spymicmac.tools.join_hexagon:main',
              'move_bad_tapas = spymicmac.tools.move_bad_tapas:main',
              'register_ortho = spymicmac.tools.register_ortho:main',
              'remove_crosses = spymicmac.tools.remove_crosses:main',
              'remove_measures = spymicmac.tools.remove_measures:main'
          ],
      },
      zip_safe=False)



#!/usr/bin/env python
import argparse
import spymicmac.micmac as mt


def _argparser():
    parser = argparse.ArgumentParser(description="Create id_fiducial.txt, MeasuresCamera.xml, and Tmp-SL-Glob.xml files for KH-9 Hexagon images.",
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-joined', action='store_true',
                        help='generate files for joined scene (220x460 mm) instead of half (220x230mm)')
    return parser


def main():
    parser = _argparser()
    args = parser.parse_args()
    mt.generate_measures_files(joined=args.joined)


if __name__ == "__main__":
    main()
